"""Utility functions for working with memory caching backends.

These functions are designed to integrate with a cache backend using
Django's cache framework. They handle creating caching keys unique to the
install and caching more complex data efficiently (such as the results of
iterators and large data normally too big for the cache).
"""

import hashlib
import hmac
import io
import logging
import pickle
import re
import zlib

from django.conf import settings
from django.core.cache import cache
from django.contrib.sites.models import Site
from django.utils.encoding import force_bytes

from djblets.cache.errors import MissingChunkError
from djblets.secrets.crypto import (aes_decrypt,
                                    aes_decrypt_iter,
                                    aes_encrypt,
                                    aes_encrypt_iter,
                                    get_default_aes_encryption_key)


logger = logging.getLogger(__name__)


DEFAULT_EXPIRATION_TIME = 60 * 60 * 24 * 30  # 1 month
CACHE_CHUNK_SIZE = 2 ** 20 - 1024  # almost 1M (memcached's slab limit)

# memcached key size constraint (typically 250, but leave a few bytes for the
# large data handling)
MAX_KEY_SIZE = 240

_INVALID_KEY_CHARS_RE = re.compile(r'[\x00-\x20\x7f]')

_NO_RESULTS = object()

_default_expiration = getattr(settings, 'CACHE_EXPIRATION_TIME',
                              DEFAULT_EXPIRATION_TIME)


class _CacheContext:
    """State and functions for performing a cache-related operation.

    This is created internally by :py:func:`cache_memoize` and
    :py:func:`cache_memoize_iter`, and ensures consistency in cache key
    names, timeouts, and state storage, simplifying the operations
    responsible for loading and saving complex cache data.

    Version Added:
        3.0
    """

    def __init__(self, cache, base_cache_key, expiration, compress_large_data,
                 use_encryption, encryption_key):
        """Initialize the context.

        Args:
            cache (object):
                The Django cache connection that all operations will work
                on. This is considered opaque.

            base_cache_key (str):
                The base cache key for caching operations.

                This is the value passed in to ::py:func:`cache_memoize` or
                 :py:func:`cache_memoize_iter`, and will be used when
                 constructing related cache keys.

            expiration (int):
                The expiration time in seconds for all data cached using
                this context.

            compress_large_data (bool):
                Whether large data will be compressed.

            use_encryption (bool):
                Whether to use encryption when storing or reading data.

                This defaults to ``False``, but can be forced on for all
                cached data by setting
                ``settings.DJBLETS_CACHE_FORCE_ENCRYPTION=True``.

            encryption_key (bytes or str):
                An explicit AES encryption key to use when passing
                ``use_encryption=True``.

                This defaults to the value in
                ``settings.DJBLETS_CACHE_DEFAULT_ENCRYPTION_KEY``, if set, or
                to the default AES encryption key for the server as provided by
                :py:func:`djblets.secrets.crypto.
                get_default_aes_encryption_key`.
        """
        if use_encryption is None:
            use_encryption = _get_default_use_encryption()

        if use_encryption:
            if encryption_key:
                assert isinstance(encryption_key, bytes)
            else:
                encryption_key = _get_default_encryption_key()
        else:
            encryption_key = None

        self.cache = cache
        self.expiration = expiration
        self.full_cache_key = None
        self.base_cache_key = base_cache_key
        self.use_encryption = use_encryption
        self.encryption_key = encryption_key
        self.compress_large_data = compress_large_data

        self.full_cache_key = self.make_key(base_cache_key)

    def make_key(self, key):
        """Return a full cache key from the provided base key.

        Args:
            key (str):
                The base cache key to make a full key from.

        Returns:
            str:
            The full cache key.
        """
        if key == self.base_cache_key and self.full_cache_key:
            return self.full_cache_key

        return make_cache_key(key,
                              use_encryption=self.use_encryption,
                              encryption_key=self.encryption_key)

    def make_subkey(self, suffix):
        """Return a full cache key combining the main key and a suffix.

        The key will be built in the form of :samp:`{mainkey}-{suffix}`, and
        will then be converted into a full cache key.

        Args:
            suffix (int or str):
                The suffix to append to the cache key.

        Returns:
            str:
            The full cache key in the form of ``<mainkey>-<suffix>``.
        """
        return self.make_key('%s-%s' % (self.base_cache_key, suffix))

    def prepare_value(self, full_cache_key, value):
        """Prepare a value for storage in cache.

        If using encryption, this will serialize and encrypt the value.
        Otherwise this returns the value as-is.

        Args:
            full_cache_key (str):
                The full cache key where this value will be stored.

            value (object):
                The value to cache.

        Returns:
            object:
            The prepared value.

        Raises:
            Exception:
                An error occurred pickling or encrypting the value. The
                The exception is logged and then raised as-is.
        """
        if self.use_encryption:
            # Normally Django's cache backends will handle pickling of data,
            # so what we set is what we get. That doesn't work when we encrypt
            # data, though, since the type information will be lost. Plus,
            # we need to be able to give a bytestring representation to
            # aes_encrypt().
            #
            # So we instead pickle the data before encrypting it, and then we
            # unpickle after we decrypt it later.
            try:
                value = pickle.dumps(value, protocol=0)
            except Exception as e:
                logger.error('Failed to serialize data for cache key "%s": %s',
                             full_cache_key, e)
                raise

            try:
                value = aes_encrypt(value, key=self.encryption_key)
            except Exception as e:
                logger.error('Failed to encrypt data for cache for key '
                             '%s: %s.',
                             full_cache_key, e)
                raise

        return value

    def load_value(self, key=None):
        """Load a value from cache.

        This will take care of converting the provided key to a full cache
        key, and decrypting and deserializing the value if needed.

        Args:
            key (str):
                The full cache key to load from cache.

                If not provided, the full main cache key will be used.

        Returns:
            object:
            The value from the cache, or :py:data:`_NO_RESULTS` if not found.

        Raises:
            Exception:
                An error occurred reading from cache or processing results.
                The exception is raised as-is.
        """
        if key is None:
            key = self.full_cache_key

        value = self.cache.get(key, _NO_RESULTS)

        if value is not _NO_RESULTS and self.use_encryption:
            try:
                value = aes_decrypt(value, key=self.encryption_key)
            except Exception as e:
                logger.error('Failed to decrypt data from cache for '
                             'key %s: %s.',
                             key, e)
                raise

            try:
                value = pickle.loads(value)
            except Exception as e:
                logger.warning('Failed to deserialize data from cache for '
                               'key %s: %s.',
                               key, e)
                raise

        return value

    def store_value(self, value, *, key=None, raw=False):
        """Store a value in cache.

        This will take care of preparing any values to be stored, if needed.

        Args:
            value (object):
                The value to store in cache.

            key (str, optional):
                The full cache key to save to cache.

                If not provided, the full main cache key will be used.

            raw (bool, optional):
                Whether to store the value directly without modifications.

        Raises:
            Exception:
                An error occurred preparing the value or writing to cache.
                The exception is raised as-is.
        """
        if key is None:
            key = self.full_cache_key

        if not raw:
            value = self.prepare_value(key, value)

        self.cache.set(key, value,
                       timeout=self.expiration)

    def store_many(self, items):
        """Store many items directly to cache.

        All keys should be full cache keys, and all values should be prepared
        already.

        Args:
            items (dict):
                A dictionary of base cache keys to raw values to store.

        Raises:
            Exception:
                An error occurred preparing the value or writing to cache.
                The exception is raised as-is.
        """
        self.cache.set_many(items,
                            timeout=self.expiration)


def _get_default_use_encryption():
    """Return whether encryption should be enabled by default.

    This will dynamically check whether encryption should be allowed, based
    on the ``settings.DJBLETS_CACHE_FORCE_ENCRYPTION`` setting. If not set,
    this defaults to ``False``.

    Returns:
        bool:
        Whether encryption should be used by default.
    """
    return getattr(settings, 'DJBLETS_CACHE_FORCE_ENCRYPTION', False)


def _get_default_encryption_key():
    """Return the default AES encryption key for caching.

    This will return the default encryption key used for caching, using
    ``settings.DJBLETS_CACHE_DEFAULT_ENCRYPTION_KEY`` if set to a valid
    AES encryption key, and falling back to
    :py:func:`djblets.secrets.crypto.get_default_aes_encryption_key`.

    Returns:
        bytes:
        The default encryption key.
    """
    key = getattr(settings, 'DJBLETS_CACHE_DEFAULT_ENCRYPTION_KEY', None)

    if key:
        if not isinstance(key, bytes):
            key = key.encode('utf-8')
            settings.DJBLETS_CACHE_DEFAULT_ENCRYPTION_KEY = key
    else:
        key = get_default_aes_encryption_key()

    return key


def _cache_fetch_large_data(cache_context, chunk_count):
    """Fetch large data from the cache.

    The main cache key indicating the number of chunks will be read, followed
    by each of the chunks. If any chunks are missing, a
    :py:class:`~djblets.cache.errors.MissingChunkError` will be immediately
    raised.

    The data is then combined and optionally uncompressed, and returned to
    the caller. The caller should iterate through the results using
    :py:func:`_cache_iter_large_data`.

    Version Changed:
        3.0:
        * Updated to take ``cache_context`` instead of additional arguments,
          and ``chunk_count`` to save a cache hit.
        * Added support for key/value encryption.

    Args:
        cache_context (_CacheContext):
            The caching operation context.

        chunk_count (int):
            The number of chunks to fetch.

    Returns:
        bytes:
        The complete fetched data, ready for deserializing.

    Raises:
        djblets.cache.errors.MissingChunkError:
            A chunk of data was missing. All cached data for the key is
            invalid.

        Exception:
            An error occurred reading from cache or processing data. The
            exception is raised as-is.
    """
    # Fetch all the chunks at once.
    chunk_keys = [
        cache_context.make_subkey(i)
        for i in range(chunk_count)
    ]
    chunks = cache_context.cache.get_many(chunk_keys)

    # Check that we have all the keys we expect, before we begin generating
    # values. We don't want to waste effort loading anything, and we want to
    # pass an error about missing keys to the caller up-front before we
    # generate anything.
    if len(chunks) != chunk_count:
        missing_keys = sorted(set(chunk_keys) - set(chunks.keys()))
        missing_keys_str = ', '.join(missing_keys)
        logger.debug('Cache miss for key(s): %s.',
                     missing_keys_str)

        raise MissingChunkError

    # Process all the chunks.
    data = (
        chunks[chunk_key][0]
        for chunk_key in chunk_keys
    )

    if cache_context.use_encryption:
        # We can iteratively decrypt these and build the results into a
        # single string below for optional decompression.
        data = aes_decrypt_iter(data, key=cache_context.encryption_key)

    data = b''.join(data)

    # Decompress them all at once, instead of streaming the results. It's
    # faster for any reasonably-sized data in cache. We'll stream depickles
    # instead.
    if cache_context.compress_large_data:
        data = zlib.decompress(data)

    return data


def _cache_iter_large_data(cache_context, data):
    """Iterate through large data that was fetched from the cache.

    This will unpickle the large data previously fetched through
    :py:func:`_cache_fetch_large_data`, and yield each object to the caller.

    Version Changed:
        3.0:
        * Updated to take ``cache_context`` instead of additional arguments.
        * Added support for key/value encryption.

    Args:
        cache_context (_CacheContext):
            The caching operation context.

        data (bytes):
            The combined data fetched from cache.

    Yields:
        object:
        Each value from cache.

    Raises:
        Exception:
            An error occurred processing data. The exception is logged and
            then raised as-is.
    """
    fp = io.BytesIO(data)

    try:
        # Unpickle all the items we're expecting from the cached data.
        #
        # There will only be one item in the case of old-style cache data.
        while True:
            try:
                yield pickle.load(fp)
            except EOFError:
                return
    except Exception as e:
        logger.warning('Unpickle error for cache key "%s": %s.',
                       cache_context.full_cache_key, e)
        raise


def _cache_compress_pickled_data(items):
    """Compress lists of items for storage in the cache.

    This works with generators, and will take each item in the list or
    generator of items, zlib-compress the data, and store it in a buffer.The
    item and a blob of compressed data will be yielded to the caller.

    Args:
        items (generator of tuple):
            The generator of item tuples prepared in
            :py:func:`_cache_store_items`.

            Each is in the form of:

            1. Byte string to compress
            2. Boolean indicating if the raw data represents valid data to
               ultimately yield to the caller in :py:func:`cache_memoize_iter`
            3. Raw data to yield back to the caller

    Yields:
        tuple:
        An item tuple, but with the first entry containing compressed data.

        Each is in the form of:

        1. Byte string containing compressed data
        2. Boolean indicating if the raw data represents valid data to
           ultimately yield to the caller in :py:func:`cache_memoize_iter`
        3. Raw data to yield back to the caller
    """
    compressor = zlib.compressobj()

    for data, has_item, item in items:
        yield compressor.compress(data), has_item, item

    remaining = compressor.flush()

    if remaining:
        yield remaining, False, None


def _cache_encrypt_data(items, encryption_key):
    """Iteratively encrypt data from items tuples.

    Version Added:
        3.0

    Args:
        items (generator of tuple):
            The generator of item tuples prepared in
            :py:func:`_cache_store_items`.

            Each is in the form of:

            1. Byte string to encrypt
            2. Boolean indicating if the raw data represents valid data to
               ultimately yield to the caller in :py:func:`cache_memoize_iter`
            3. Raw data to yield back to the caller

        encryption_key (bytes, optional):
            The AES encryption key to use.

    Yields:
        tuple:
        Each is in the form of:

        1. Byte string containing encrypted data
        2. Boolean indicating if the raw data represents valid data to
           ultimately yield to the caller in :py:func:`cache_memoize_iter`
        3. Raw data to yield back to the caller
    """
    # We're working with a generator yielding tuples containing items.
    # The first entry in each tuple is a piece of the data we want to
    # encrypt. We want to encrypt iteratively, and we don't want to have
    # to load the entire results of the generator into memory at once.
    #
    # So we have to feed the first index of each tuple iteratively into
    # aes_encrypt_iter as a generator, but we also want to retain the other
    # data in each tuple and yield it along with the encrypted results.
    #
    # There isn't a one-to-one mapping between what goes in and what comes
    # out, particularly when dealing with compression.
    #
    # So here's the plan of attack:
    #
    # 1. We feed data into aes_encrypt_iter(), and keep a queue of the item
    #    data (has_item and item).
    #
    # 2. As we receive results, we pull off the front of the queue and yield
    #    it along with the data.
    #
    #    We then go through anything else in the queue (one item coming
    #    in could result in multiple coming out). Those yield those
    #    values, but without any data scheduled for storage, just scheduled
    #    to yield as results.
    #
    # 3. If there was nothing in the queue when we get data, then yield the
    #    data but with has_item=False, so it'll be stored but not yielded as
    #    results.
    item_queue = []

    def _gen_data():
        for data, has_item, item in items:
            item_queue.append((has_item, item))
            yield data

    for data in aes_encrypt_iter(_gen_data(),
                                 key=encryption_key):
        if item_queue:
            # There were items in the queue. Yield the first with the data,
            # and anything else without (those will be returned as results
            # but won't be stored in cache).
            yield (data,) + item_queue[0]

            for item in item_queue[1:]:
                yield (None,) + item
        else:
            # The queue is empty. Yield what data we have for caching, but
            # don't yield for results.
            yield data, False, None

        item_queue.clear()


def _cache_store_chunks(cache_context, items):
    """Store a list of items as chunks in the cache.

    The list of items will be combined into chunks and stored in the
    cache as efficiently as possible. Each item in the list will be
    yielded to the caller as it's fetched from the list or generator.

    Version Changed:
        3.0:
        * Updated to take ``cache_context`` instead of additional arguments.
        * Added support for encrypting keys and data through the
          ``use_encryption`` and ``encryption_key`` arguments.

    Args:
        cache_context (_CacheContext):
            The caching operation context.

        items (generator of tuple):
            The generator of item tuples prepared in
            :py:func:`_cache_store_items`.

            Each is in the form of:

            1. Byte string to store in cache
            2. Boolean indicating if the raw data represents valid data to
               ultimately yield to the caller in :py:func:`cache_memoize_iter`
            3. Raw data to yield back to the caller

    Yields:
        object:
        Each chunk of original, unmodified item data being cached.
    """
    chunks_data = io.BytesIO()
    chunks_data_len = 0
    read_start = 0
    i = 0

    for data, has_item, item in items:
        if has_item:
            yield item

        if not data:
            continue

        chunks_data.write(data)
        chunks_data_len += len(data)

        if chunks_data_len > CACHE_CHUNK_SIZE:
            # We have enough data to fill a chunk now. Start processing
            # what we've stored and create cache keys for each chunk.
            # Anything remaining will be stored for the next round.
            chunks_data.seek(read_start)
            cached_data = {}

            while chunks_data_len > CACHE_CHUNK_SIZE:
                chunk = chunks_data.read(CACHE_CHUNK_SIZE)
                chunk_len = len(chunk)
                chunks_data_len -= chunk_len
                read_start += chunk_len

                # Note that we wrap the chunk in a list so that the cache
                # backend won't try to perform any conversion on the string.
                cached_data[cache_context.make_subkey(i)] = [chunk]
                i += 1

            # Store the keys in the cache in a single request.
            cache_context.store_many(cached_data)

            # Reposition back at the end of the stream.
            chunks_data.seek(0, io.SEEK_END)

    if chunks_data_len > 0:
        # There's one last bit of data to store. Note that this should be
        # less than the size of a chunk,
        assert chunks_data_len <= CACHE_CHUNK_SIZE

        chunks_data.seek(read_start)
        chunk = chunks_data.read()
        cache_context.store_value([chunk],
                                  key=cache_context.make_subkey(i),
                                  raw=True)
        i += 1

    # Store the final count.
    cache_context.store_value('%d' % i)


def _cache_store_items(cache_context, items):
    """Store items in the cache.

    The items will be individually pickled and combined into a binary blob,
    which can then optionally be compressed. The resulting data is then
    cached over one or more keys, each representing a chunk about 1MB in size.

    A main cache key will be set that contains information on the other keys.

    Version Changed:
        3.0:
        * Updated to take ``cache_context`` instead of additional arguments.
        * Added support for encrypting keys and data through the
          ``use_encryption`` and ``encryption_key`` arguments.

    Args:
        cache_context (_CacheContext):
            The caching operation context.

        items (generator):
            The generator of data to store.

    Yields:
        tuple:
        Each is in the form of:

        1. Byte string to store in the cache
        2. Boolean indicating if the raw data represents valid data to
           ultimately yield to the caller in :py:func:`cache_memoize_iter`
        3. Raw data to yield back to the caller
    """
    # Note that we want to use pickle protocol 0 in order to be compatible
    # across both Python 2 and 3. On Python 2, 0 is the default.
    results = (
        (pickle.dumps(item, protocol=0), True, item)
        for item in items
    )

    if cache_context.compress_large_data:
        results = _cache_compress_pickled_data(results)

    if cache_context.use_encryption:
        # Note that the order of operations here is important. We must
        # compress and *then* encrypt. Encrypted AES data cannot be
        # compressed.
        results = _cache_encrypt_data(
            results,
            encryption_key=cache_context.encryption_key)

    yield from _cache_store_chunks(cache_context=cache_context,
                                   items=results)


def cache_memoize_iter(key, items_or_callable,
                       expiration=_default_expiration,
                       force_overwrite=False,
                       compress_large_data=True,
                       use_encryption=None,
                       encryption_key=None):
    """Memoize an iterable list of items inside the configured cache.

    If the provided list of items is a function, the function must return a
    an iterable object, such as a list or a generator.

    If a generator is provided, directly or through a function, then each
    item will be immediately yielded to the caller as they're retrieved, and
    the cached entries will be built up as the items are processed.

    The data is assumed to be big enough that it must be pickled,
    optionally compressed, and stored as chunks in the cache.

    Data can be encrypted using AES encryption, for safe storage of potentially
    sensitive information or state. This adds to the processing time slightly,
    but can be important for cache keys containing sensitive information or
    that impact access control, particularly in the event that the cache is
    compromised or shared between services.

    The result from this function is always a generator. Note that it's
    important that the generator be allowed to continue until completion, or
    the data won't be retrievable from the cache.

    Version Changed:
        3.0:
        Added support for encrypting keys and data through the
        ``use_encryption`` and ``encryption_key`` arguments.

    Args:
        key (unicode):
            The key to use in the cache.

        items_or_callable (list or callable):
            A list of items or callable returning a list of items to cache,
            if the key is not already found in cache.

        expiration (int):
            The expiration time for the key, in seconds.

        force_overwrite (bool):
            If ``True``, the value will always be computed and stored
            regardless of whether it exists in the cache already.

        compress_large_data (bool):
            If ``True``, the data will be zlib-compressed.

        use_encryption (bool, optional):
            Whether to use encryption when storing or reading data.

            If reading data, and if the data cannot be decrypted with the
            given key, then the data will be considered to have fallen out
            of cache.

            This defaults to ``False``, but can be forced on for all cached
            data by setting ``settings.DJBLETS_CACHE_FORCE_ENCRYPTION=True``.

            Version Added:
                3.0

        encryption_key (bytes or str, optional):
            An explicit AES encryption key to use when passing
            ``use_encryption=True``.

            This defaults the value in
            ``settings.DJBLETS_CACHE_DEFAULT_ENCRYPTION_KEY``, if set, or to
            the default AES encryption key for the server as provided by
            :py:func:`djblets.secrets.crypto.get_default_aes_encryption_key`.

            Version Added:
                3.0

    Yields:
        object:
        The list of items from the cache or from ``items_or_callable`` if
        uncached.
    """
    cache_context = _CacheContext(
        cache=cache,
        base_cache_key=key,
        expiration=expiration,
        compress_large_data=compress_large_data,
        use_encryption=use_encryption,
        encryption_key=encryption_key)
    full_cache_key = cache_context.full_cache_key

    results = _NO_RESULTS

    if not force_overwrite:
        chunk_count = cache_context.load_value()

        if chunk_count is not _NO_RESULTS:
            try:
                results = _cache_iter_large_data(
                    cache_context=cache_context,
                    data=_cache_fetch_large_data(
                        cache_context=cache_context,
                        chunk_count=int(chunk_count)))
            except Exception as e:
                logger.warning('Failed to fetch large data from cache for '
                               'key "%s": %s',
                               full_cache_key, e)
                results = _NO_RESULTS
        else:
            logger.debug('Cache miss for key "%s"' % full_cache_key)

    if results is _NO_RESULTS:
        if callable(items_or_callable):
            items = items_or_callable()
        else:
            items = items_or_callable

        results = _cache_store_items(cache_context=cache_context,
                                     items=items)

    yield from results


def cache_memoize(key,
                  lookup_callable,
                  expiration=_default_expiration,
                  force_overwrite=False,
                  large_data=False,
                  compress_large_data=True,
                  use_generator=False,
                  use_encryption=None,
                  encryption_key=None):
    """Memoize the results of a callable inside the configured cache.

    Data can be encrypted using AES encryption, for safe storage of potentially
    sensitive information or state. This adds to the processing time slightly,
    but can be important for cache keys containing sensitive information or
    that impact access control, particularly in the event that the cache is
    compromised or shared between services.

    Version Changed:
        3.0:
        Added support for encrypting keys and data through the
        ``use_encryption`` and ``encryption_key`` arguments.

    Version Changed:
        2.2.4:
        Added support for non-iterable value types.

    Args:
        key (unicode):
            The key to use in the cache.

        lookup_callable (callable):
            A callable to execute in the case where the data did not exist in
            the cache.

        expiration (int):
            The expiration time for the key, in seconds.

        force_overwrite (bool):
            If ``True``, the value will always be computed and stored
            regardless of whether it exists in the cache already.

        large_data (bool):
            If ``True``, the resulting data will be pickled, gzipped, and
            (potentially) split up into megabyte-sized chunks. This is useful
            for very large, computationally intensive hunks of data which we
            don't want to store in a database due to the way things are
            accessed.

        compress_large_data (bool):
            Compresses the data with zlib compression when ``large_data``
            is ``True``.

        use_generator (bool, deprecated):
            This parameter is no longer used and will be removed in Djblets
            3.0.

        use_encryption (bool, optional):
            Whether to use encryption when storing or reading data.

            If reading data, and if the data cannot be decrypted with the
            given key, then the data will be considered to have fallen out
            of cache.

            This defaults to ``False``, but can be forced on for all cached
            data by setting ``settings.DJBLETS_CACHE_FORCE_ENCRYPTION=True``.

            Version Added:
                3.0

        encryption_key (bytes or str, optional):
            An explicit AES encryption key to use when passing
            ``use_encryption=True``.

            This defaults the value in
            ``settings.DJBLETS_CACHE_DEFAULT_ENCRYPTION_KEY``, if set, or to
            the default AES encryption key for the server as provided by
            :py:func:`djblets.secrets.crypto.get_default_aes_encryption_key`.

            Version Added:
                3.0

    Returns:
        object:
        The cached data, or the result of ``lookup_callable`` if uncached.
    """
    if large_data:
        results = list(cache_memoize_iter(
            key,
            lambda: [lookup_callable()],
            expiration=expiration,
            force_overwrite=force_overwrite,
            compress_large_data=compress_large_data,
            use_encryption=use_encryption,
            encryption_key=encryption_key))

        assert len(results) == 1

        return results[0]
    else:
        cache_context = _CacheContext(
            cache=cache,
            base_cache_key=key,
            expiration=expiration,
            compress_large_data=compress_large_data,
            use_encryption=use_encryption,
            encryption_key=encryption_key)
        full_cache_key = cache_context.full_cache_key

        if not force_overwrite:
            try:
                result = cache_context.load_value(full_cache_key)
            except Exception:
                # We've already logged this. Proceed to generate new data.
                pass

            if result is not _NO_RESULTS:
                return result
            else:
                logger.debug('Cache miss for key "%s"' % full_cache_key)

        data = lookup_callable()

        # Most people will be using memcached, and memcached has a limit of
        # 1MB. Data this big should be broken up somehow, so let's warn
        # about this. Users should hopefully be using large_data=True which
        # will handle this appropriately.
        #
        # If we do get here, we try to do some sanity checking.
        # python-memcached will return a result in the case where the data
        # exceeds the value size, which Django will then silently use to clear
        # out the key. We won't know at all whether we had success unless we
        # come back and try to verify the value.
        #
        # This check handles the common case of large string data being stored
        # in cache. It's still possible to attempt to store large data
        # structures (where len(data) might be something like '6' but the
        # serialized value is huge), where this can still fail.
        if (isinstance(data, str) and
            len(data) >= CACHE_CHUNK_SIZE):
            logger.warning('Cache data for key "%s" (length %s) may be too '
                           'big for the cache.',
                           full_cache_key, len(data))

        try:
            cache_context.store_value(data, key=full_cache_key)
        except Exception as e:
            logger.error('Unable to store cached data in key "%s": %s',
                         full_cache_key, e)

        return data


def make_cache_key(key, use_encryption=None, encryption_key=None):
    """Create a cache key guaranteed to avoid conflicts and size limits.

    The cache key will be prefixed by the site's domain, and will be
    changed to an SHA256 hash if it's larger than the maximum key size or
    contains characters not compatible with the cache backend.

    Version Changed:
        3.0:
        * Added support for encrypting keys through the
          ``use_encryption`` and ``encryption_key`` arguments.
        * Changed the hash format for keys to use SHA256 instead of MD5.
          This will invalidate all old keys in cache, but reduces chances
          of collision.
        * Keys will automatically use the hash format if they contain
          characters unsupported by the cache backend.
        * The return type is now :py:class:`str`, to generate keys that are
          more suitable for modern versions of Django.

    Args:
        key (str):
            The base key to generate a cache key from.

        use_encryption (bool, optional):
            Whether to generate an encrypted key.

            This will generate a HMAC digest of the key. There will be no
            identifying information in the resulting key.

            This defaults to ``False``, but can be forced on for all cached
            data by setting ``settings.DJBLETS_CACHE_FORCE_ENCRYPTION=True``.

            Version Added:
                3.0

        encryption_key (bytes, optional):
            An explicit AES encryption key to use when passing
            ``use_encryption=True``.

            This defaults the value in
            ``settings.DJBLETS_CACHE_DEFAULT_ENCRYPTION_KEY``, if set, or to
            the default AES encryption key for the server as provided by
            :py:func:`djblets.secrets.crypto.get_default_aes_encryption_key`.

            Version Added:
                3.0

    Returns:
        str:
        A cache key suitable for use with the cache backend.
    """
    if use_encryption is None:
        use_encryption = _get_default_use_encryption()

    try:
        site = Site.objects.get_current()

        # The install has a Site app, so prefix the domain to the key.
        # If a SITE_ROOT is defined, also include that, to allow for multiple
        # instances on the same host.
        site_root = getattr(settings, 'SITE_ROOT', None)

        if site_root:
            key = '%s:%s:%s' % (site.domain, site_root, key)
        else:
            key = '%s:%s' % (site.domain, key)
    except Exception:
        # The install doesn't have a Site app, so use the key as-is.
        pass

    if use_encryption:
        if encryption_key:
            assert isinstance(encryption_key, bytes)
        else:
            encryption_key = _get_default_encryption_key()

        # Construct a HMAC digest of the key using the encryption key.
        # The result will be a SHA256 hash.
        key = hmac.new(
            encryption_key,
            msg=force_bytes(key),
            digestmod=hashlib.sha256).hexdigest()
    else:
        # Normalize any invalid characters in the key.
        key = _INVALID_KEY_CHARS_RE.sub(lambda m: '\\x%02x' % ord(m.group(0)),
                                        key)

        if len(key) > MAX_KEY_SIZE:
            digest = hashlib.sha256(key.encode('utf-8')).hexdigest()

            # Replace the excess part of the key with a digest of the key
            key = key[:MAX_KEY_SIZE - len(digest)] + digest

    return key



"""Administration UI registrations for site configurations."""

from django.contrib import admin

from djblets.siteconfig.models import SiteConfiguration


class SiteConfigurationAdmin(admin.ModelAdmin):
    list_display = ('site', 'version')


admin.site.register(SiteConfiguration, SiteConfigurationAdmin)



import functools
import re

from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from django.db.models.manager import Manager
from django.db.models.query_utils import Q


class LocalDataQuerySet(object):
    """A QuerySet that operates on generic data provided by the caller.

    This can be used in some circumstances when code requires a QuerySet,
    but where the data doesn't come from the database. The caller can
    instantiate one of these and provide it.

    This doesn't perform full support for all of QuerySet's abilities. It
    does, however, support the following basic functions:

    * all
    * clone
    * count
    * exclude
    * filter
    * get
    * order_by
    * prefetch_related
    * select_related

    As well as the operators expected by consumers of QuerySet, such as
    __len__ and __iter__.

    This is particularly handy with WebAPIResource.
    """

    PROPERTY_REGEX = re.compile(r'(-?)(\w+)')

    def __init__(self, data):
        self._data = data

    def all(self):
        """Returns a cloned copy of this queryset."""
        return self.clone()

    def clone(self):
        """Returns a cloned copy of this queryset."""
        return LocalDataQuerySet(list(self._data))

    def count(self):
        """Returns the number of items in this queryset."""
        return len(self._data)

    def exclude(self, **kwargs):
        """Returns a queryset excluding items from this queryset.

        The result will be a LocalDataQuerySet that contains all items from
        this queryset that do not contain attributes with values matching
        those that were passed to this function as keyword arguments.
        """
        return LocalDataQuerySet(
            list(self._filter_or_exclude(return_matches=False, **kwargs)))

    def filter(self, **kwargs):
        """Returns a queryset filtering items from this queryset.

        The result will be a LocalDataQuerySet that contains all items from
        this queryset that contain attributes with values matching those that
        were passed to this function as keyword arguments.
        """
        return LocalDataQuerySet(
            list(self._filter_or_exclude(return_matches=True, **kwargs)))

    def get(self, **kwargs):
        """Returns a single result from this queryset.

        This will return a single result from the list of items in this
        queryset. If keyword arguments are provided, they will be used
        to filter the queryset down.

        There must be only one item in the queryset matching the given
        criteria, or a MultipleObjectsReturned will be raised. If there are
        no items, then an ObjectDoesNotExist will be raised.
        """
        clone = self.filter(**kwargs)
        count = len(clone)

        if count == 1:
            return clone[0]
        elif count == 0:
            raise ObjectDoesNotExist('%s matching query does not exist.'
                                     % self._data.__class__.__name__)
        else:
            raise MultipleObjectsReturned(
                'get() returned more than one %s -- it returned %s!'
                % (self._data.__class__.__name__, count))

    def order_by(self, *attrs, **kwargs):
        """Returns a queryset ordering items by the specified attributes.

        The result will be a LocalDataQuerySet that contains all items from
        this queryset ordered by the attributes specified. If multiple
        attributes are specified, the items are sorted by the first attribute
        and ties are broken by the other following attributes.

        All items are sorted in ascending order. To specify a descending
        order, an attribute must have a '-' prepended to the name, such as
        `-attribute_A`.
        """
        def compare(element_a, element_b):
            """Compare two elements of the queryset."""
            for attr in attrs:
                match = self.PROPERTY_REGEX.search(attr)
                attr = match.group(2)

                # Set the default result based on whether this is in ascending
                # or descending order.
                if match.group(1):
                    direction = -1
                else:
                    direction = 1

                attr_a = getattr(element_a, attr)
                attr_b = getattr(element_b, attr)

                if attr_a < attr_b:
                    return -direction
                elif attr_a > attr_b:
                    return direction

            return 0

        copy = self.clone()
        copy._data.sort(key=functools.cmp_to_key(compare))
        return copy

    def prefetch_related(self, *args, **kwargs):
        """Stub for compatibility with QuerySet.prefetch_related.

        This will simply return a clone of this queryset.
        """
        return self.clone()

    def select_related(self, *args, **kwargs):
        """Stub for compatibility with QuerySet.select_related.

        This will simply return a clone of this queryset.
        """
        return self.clone()

    def __contains__(self, i):
        return i in self._data

    def __getitem__(self, i):
        return self._data[i]

    def __getslice__(self, i, j):
        return self._data[i:j]

    def __iter__(self):
        for i in self._data:
            yield i

    def __len__(self):
        return len(self._data)

    def _filter_or_exclude(self, return_matches=True, **kwargs):
        for item in self:
            match = True

            for key, value in kwargs.items():
                if getattr(item, key) != value:
                    match = False
                    break

            if match == return_matches:
                yield item


def get_object_or_none(cls, *args, **kwargs):
    """Return a model instance or None if one can not be found.

    Args:
        cls (type or django.db.models.manager.Manager):
            Either the model, a subclass of :py:class:`django.db.models.Model`,
            or its manager.

        *args (tuple):
            Arguments to pass to
            :py:meth:`~django.db.models.manager.Manger.get`.

        **kwargs (dict):
            Keyword arguments to pass to
            :py:meth:`~django.db.models.manager.Manger.get`.

    Returns:
        django.db.models.Model:
        The model instance, if it could be found, or :py:data:`None` otherwise.
    """
    if isinstance(cls, Manager):
        manager = cls
        cls = manager.model
    else:
        manager = cls._default_manager

    try:
        return manager.get(*args, **kwargs)
    except cls.DoesNotExist:
        return None


def prefix_q(prefix, q, clone=True):
    """Prefix a query expression.

    A :py:class:`query expression <django.db.models.query_utils.Q>` is used in
    :py:meth:`Manager.get <django.db.models.manager.Manager.get>` and
    :py:meth:`Manager.filter <django.db.models.manager.Manager.filter>`
    methods. These expressions are made of a tree of key-value pairs.

    Prefixing these expressions allows a query expression for one model to be
    used in a related model. The resulting prefixed expression will have every
    key in every key-value pair prefixed with the given prefix.

    By default, the query expression will be cloned. That is, the given query
    expression will not be modified and a prefixed copy will be returned.

    Args:
        prefix (unicode):
            The prefix to add to each member in the expression. This should be
            the name of a field and will be automatically suffixed by ``__``.

        q (django.db.models.query_utils.Q):
            The expression to prefix.

        clone (bool):
            Determine if the query should be cloned. If this is ``False``,
            ``q`` will be modified in place. This defaults to ``True``.

    Returns:
        django.db.models.query_utils.Q:
        The prefixed query expression.

    Example usage:
        .. code-block:: python

           from django.db import models
           from django.db.query_utils import Q
           from djblets.db.query import prefix_q

           # Given the following models:
           class A(models.Model):
               foo = models.IntegerField()
               bar = models.IntegerField()

           class B(models.Model):
               fk = models.ForeignKey(A, on_delete=models.CASCADE)
               baz = models.IntegerField()

           # And the following query expressions:
           a_q = Q(foo=1) & Q(bar=2)
           b_q = Q(fk__foo=1) & Q(fk__bar=2)

           # The following queries are equivalent:
           B.objects.filter(b_q)
           B.objects.filter(prefix_q('fk', a_q))
    """
    if clone:
        q = _clone_q(q)

    for i, child in enumerate(q.children):
        # django.utils.tree (which Q inherits from) stores its children as a
        # list. Each member is either an instance of Q (in the case of a nested
        # expression) or a tuple of the form (key_name, key_val).
        if isinstance(child, Q):
            prefix_q(prefix, child, False)
        else:
            assert type(q.children[i]) is tuple

            key, value = q.children[i]

            # This is going to be pure ASCII data (it's basically keyword
            # argument names). We'll use the native string type on Python.
            q.children[i] = (str('%s__%s' % (prefix, key)),
                             value)

    return q


def _clone_q(q):
    """Clone a query expression.

    This will clone a :py:class:`~django.db.models.Q` object and all children,
    allowing for modifications to be made on its tree.

    Args:
        q (django.db.models.Q):
            The query expression.

    Returns:
        django.db.models.Q:
        A clone of the query expression.
    """
    if hasattr(q, 'clone'):
        q = q.clone()
    else:
        # Newer versions of Django do not support this, so we have to
        # re-implement the logic.
        new_q = q.__class__._new_instance(children=[],
                                          connector=q.connector,
                                          negated=q.negated)

        for child in q.children:
            if isinstance(child, Q):
                new_child = _clone_q(child)
            elif hasattr(child, 'clone'):
                new_child = child.clone()
            else:
                new_child = child

            new_q.children.append(new_child)

    return q



import time
from datetime import datetime

from django.core.cache import cache
from djblets.cache.backend import make_cache_key


class GenerationSynchronizer(object):
    """Manages the synchronization of generation state across processes.

    This is a utility class that makes it easy for consumers to synchronize
    a generation ID between processes and servers. This ID can be used to
    check whether a process has the latest version of some data, updating
    if the version it has is expired.

    Callers should create a :py:class:`GenerationSynchronizer` with a stable
    cache key, which will be used for communication between processes/servers.
    The initial generation number will be fetched, or created if one does not
    already exist.

    When the caller has updated something in the state, it must call
    :py:meth:`mark_updated`. This will bump the synchronization generation
    number, which will invalidate other processes.

    Other callers, upon noticing that their state is expired (through
    :py:meth:`is_expired`) can re-fetch or re-compute the data needed and
    then call :py:meth:`refresh` to refresh the instance's counter from the
    cache.

    Attributes:
        sync_gen (int):
            The synchronization generation number last fetched or set by
            this instance.

        cache_key (unicode):
            The synchronization cache key.
    """

    def __init__(self, cache_key, normalize_cache_key=True):
        """Initialize the synchronizer.

        Args:
            cache_key (unicode):
                The base cache key used for all synchronization. This will be
                normalized by
                :py:func:`~djblets.cache.backends.make_cache_key`.

            normalize_cache_key (bool, optional):
                Whether to normalize the cache key. Normalizing it will
                ensure it can fit within the key length constraints, and
                reduces changes of colliding with keys from other services.
                This is enabled by default.
        """
        if normalize_cache_key:
            cache_key = make_cache_key(cache_key)

        self.cache_key = cache_key
        self.sync_gen = None

        self._fetch_or_create_sync_gen()

    def is_expired(self):
        """Return whether the current state has expired.

        Returns:
            bool:
            ``True`` if the state has expired. ``False`` if this has the
            latest cached generation.
        """
        sync_gen = self._get_latest_sync_gen()

        return (sync_gen is None or
                (type(sync_gen) is int and sync_gen != self.sync_gen))

    def refresh(self):
        """Refresh the generation ID from cache.

        This should be called after having updated to the latest state.
        """
        self._fetch_or_create_sync_gen()

    def clear(self):
        """Clear the cached generation ID.

        This will expire all existing caches and force all processes to
        re-fetch and store the cache state.
        """
        cache.delete(self.cache_key)

    def mark_updated(self):
        """Mark the synchronized state as having been updated.

        All other processes will find their state expired, and will need to
        re-update.
        """
        try:
            self._increment_sync_gen()
        except ValueError:
            self._fetch_or_create_sync_gen()

    def _increment_sync_gen(self):
        """Increment the synchronization generation ID."""
        self.sync_gen = cache.incr(self.cache_key)

    def _fetch_or_create_sync_gen(self):
        """Return or create a new synchronization generation ID.

        If one already exists in the cache, it will be returned. Otherwise,
        the a new generation ID will be stored.
        """
        sync_gen = int(time.mktime(datetime.now().timetuple()))

        if cache.add(self.cache_key, sync_gen):
            self.sync_gen = sync_gen
        else:
            self.sync_gen = self._get_latest_sync_gen()

    def _get_latest_sync_gen(self):
        """Return the latest synchronization generation ID.

        Returns:
            int: The latest generation ID from cache.
        """
        return cache.get(self.cache_key)



from django import forms
from kgb import SpyAgency

from djblets.conditions.choices import BaseConditionChoice, ConditionChoices
from djblets.conditions.conditions import Condition, ConditionSet
from djblets.conditions.errors import (ConditionChoiceNotFoundError,
                                       ConditionOperatorNotFoundError,
                                       InvalidConditionModeError,
                                       InvalidConditionValueError)
from djblets.conditions.operators import (BaseConditionOperator,
                                          ConditionOperators)
from djblets.conditions.values import ConditionValueFormField
from djblets.testing.testcases import TestCase


class BasicTestOperator(BaseConditionOperator):
    operator_id = 'basic-test-op'


class BooleanTestOperator(BaseConditionOperator):
    operator_id = 'boolean-test-op'
    value_field = ConditionValueFormField(forms.BooleanField())


class NoValueTestOperator(BaseConditionOperator):
    operator_id = 'no-value-test-op'
    value_field = None


class EqualsTestOperator(BaseConditionOperator):
    operator_id = 'equals-test-op'

    def matches(self, match_value, condition_value):
        return match_value == condition_value


class BasicTestChoice(BaseConditionChoice):
    choice_id = 'basic-test-choice'
    operators = ConditionOperators([BasicTestOperator])
    default_value_field = ConditionValueFormField(forms.CharField())


class EqualsTestChoice(BaseConditionChoice):
    choice_id = 'equals-test-choice'
    operators = ConditionOperators([EqualsTestOperator])
    default_value_field = ConditionValueFormField(forms.CharField())


class ConditionTests(SpyAgency, TestCase):
    """Unit tests for djblets.conditions.conditions.Condition."""

    def test_deserialize(self):
        """Testing Condition.deserialize"""
        choices = ConditionChoices([BasicTestChoice])

        condition = Condition.deserialize(
            choices,
            {
                'choice': 'basic-test-choice',
                'op': 'basic-test-op',
                'value': 'my-value',
            })

        self.assertEqual(condition.choice.__class__, BasicTestChoice)
        self.assertEqual(condition.operator.__class__, BasicTestOperator)
        self.assertEqual(condition.value, 'my-value')
        self.assertEqual(condition.raw_value, 'my-value')

    def test_deserialize_with_choice_kwargs(self):
        """Testing Condition.deserialize with choice_kwargs"""
        choices = ConditionChoices([BasicTestChoice])

        condition = Condition.deserialize(
            choices,
            {
                'choice': 'basic-test-choice',
                'op': 'basic-test-op',
                'value': 'my-value',
            },
            choice_kwargs={
                'abc': 123,
            })

        self.assertEqual(condition.choice.__class__, BasicTestChoice)
        self.assertEqual(condition.choice.extra_state, {'abc': 123})
        self.assertEqual(condition.operator.__class__, BasicTestOperator)
        self.assertEqual(condition.value, 'my-value')
        self.assertEqual(condition.raw_value, 'my-value')

    def test_deserialize_with_op_value_field(self):
        """Testing Condition.deserialize with operator's value_field"""
        class MyChoice(BaseConditionChoice):
            choice_id = 'my-choice'
            operators = ConditionOperators([BooleanTestOperator])
            default_value_field = ConditionValueFormField(forms.CharField())

        choices = ConditionChoices([MyChoice])

        condition = Condition.deserialize(
            choices,
            {
                'choice': 'my-choice',
                'op': 'boolean-test-op',
                'value': True,
            })

        self.assertEqual(condition.choice.__class__, MyChoice)
        self.assertEqual(condition.operator.__class__, BooleanTestOperator)
        self.assertEqual(condition.value, True)
        self.assertEqual(condition.raw_value, True)

    def test_deserialize_with_no_value_field(self):
        """Testing Condition.deserialize with no value_field"""
        class MyChoice(BaseConditionChoice):
            choice_id = 'my-choice'
            operators = ConditionOperators([NoValueTestOperator])
            default_value_field = ConditionValueFormField(forms.CharField())

        choices = ConditionChoices([MyChoice])

        condition = Condition.deserialize(
            choices,
            {
                'choice': 'my-choice',
                'op': 'no-value-test-op',
                'value': True,
            })

        self.assertEqual(condition.choice.__class__, MyChoice)
        self.assertEqual(condition.operator.__class__, NoValueTestOperator)
        self.assertEqual(condition.value, None)
        self.assertEqual(condition.raw_value, None)

    def test_deserialize_with_missing_choice(self):
        """Testing Condition.deserialize with missing choice in data"""
        choices = ConditionChoices()

        with self.assertRaises(ConditionChoiceNotFoundError) as cm:
            Condition.deserialize(
                choices,
                {
                    'op': 'my-op',
                    'value': 'my-value',
                },
                condition_index=1)

        e = cm.exception
        self.assertEqual(str(e), 'A choice is required.')
        self.assertEqual(e.condition_index, 1)

    def test_deserialize_with_missing_operator(self):
        """Testing Condition.deserialize with missing operator in data"""
        choices = ConditionChoices()

        with self.assertRaises(ConditionOperatorNotFoundError) as cm:
            Condition.deserialize(
                choices,
                {
                    'choice': 'my-choice',
                    'value': 'my-value',
                },
                condition_index=1)

        e = cm.exception
        self.assertEqual(str(e), 'An operator is required.')
        self.assertEqual(e.condition_index, 1)

    def test_deserialize_with_missing_value(self):
        """Testing Condition.deserialize with missing value in data"""
        choices = ConditionChoices([BasicTestChoice])

        with self.assertRaises(InvalidConditionValueError) as cm:
            Condition.deserialize(
                choices,
                {
                    'choice': 'basic-test-choice',
                    'op': 'basic-test-op',
                },
                condition_index=1)

        e = cm.exception
        self.assertEqual(str(e), 'A value is required.')
        self.assertEqual(e.condition_index, 1)

    def test_deserialize_with_invalid_choice(self):
        """Testing Condition.deserialize with invalid choice in data"""
        choices = ConditionChoices()

        with self.assertRaises(ConditionChoiceNotFoundError) as cm:
            Condition.deserialize(
                choices,
                {
                    'choice': 'invalid-choice',
                    'op': 'my-op',
                    'value': 'my-value',
                },
                condition_index=1)

        e = cm.exception
        self.assertEqual(str(e),
                         'No condition choice was found matching '
                         '"invalid-choice".')
        self.assertEqual(e.choice_id, 'invalid-choice')
        self.assertEqual(e.condition_index, 1)

    def test_deserialize_with_invalid_operator(self):
        """Testing Condition.deserialize with invalid operator in data"""
        class MyChoice(BaseConditionChoice):
            choice_id = 'my-choice'
            operators = ConditionOperators()

        choices = ConditionChoices([MyChoice])

        with self.assertRaises(ConditionOperatorNotFoundError) as cm:
            Condition.deserialize(
                choices,
                {
                    'choice': 'my-choice',
                    'op': 'invalid-op',
                    'value': 'my-value',
                },
                condition_index=1)

        e = cm.exception
        self.assertEqual(str(e),
                         'No operator was found matching '
                         '"invalid-op".')
        self.assertEqual(e.operator_id, 'invalid-op')
        self.assertEqual(e.condition_index, 1)

    def test_deserialize_with_invalid_value(self):
        """Testing Condition.deserialize with invalid value in data"""
        class MyChoice(BaseConditionChoice):
            choice_id = 'my-choice'
            operators = ConditionOperators([BasicTestOperator])
            default_value_field = ConditionValueFormField(forms.IntegerField())

        choices = ConditionChoices([MyChoice])

        with self.assertRaises(InvalidConditionValueError) as cm:
            Condition.deserialize(
                choices,
                {
                    'choice': 'my-choice',
                    'op': 'basic-test-op',
                    'value': 'invalid-value'
                },
                condition_index=1)

        e = cm.exception
        self.assertEqual(str(e), 'Enter a whole number.')
        self.assertEqual(e.code, 'invalid')
        self.assertEqual(e.condition_index, 1)

    def test_serialize(self):
        """Testing Condition.serialize"""
        class MyChoice(BaseConditionChoice):
            choice_id = 'my-choice'
            operators = ConditionOperators([BasicTestOperator])
            default_value_field = ConditionValueFormField(forms.IntegerField())

        choice = MyChoice()
        condition = Condition(choice, choice.get_operator('basic-test-op'),
                              123)

        self.assertEqual(
            condition.serialize(),
            {
                'choice': 'my-choice',
                'op': 'basic-test-op',
                'value': 123,
            })

    def test_serialize_without_value_field(self):
        """Testing Condition.serialize without a value_field"""
        class MyChoice(BaseConditionChoice):
            choice_id = 'my-choice'
            operators = ConditionOperators([BasicTestOperator])

        choice = MyChoice()
        condition = Condition(choice, choice.get_operator('basic-test-op'))

        self.assertEqual(
            condition.serialize(),
            {
                'choice': 'my-choice',
                'op': 'basic-test-op',
            })

    def test_matches_with_match(self):
        """Testing Condition.matches with match"""
        choice = EqualsTestChoice()
        condition = Condition(choice, choice.get_operator('equals-test-op'),
                              'abc123')
        self.assertTrue(condition.matches('abc123'))

    def test_matches_with_no_match(self):
        """Testing Condition.matches with no match"""
        choice = EqualsTestChoice()
        condition = Condition(choice, choice.get_operator('equals-test-op'),
                              'abc123')
        self.assertFalse(condition.matches('def123'))


class ConditionSetTests(TestCase):
    """Unit tests for djblets.conditions.conditions.ConditionSet."""

    def test_deserialize(self):
        """Testing ConditionSet.deserialize"""
        choices = ConditionChoices([BasicTestChoice])

        condition_set = ConditionSet.deserialize(
            choices,
            {
                'mode': 'any',
                'conditions': [
                    {
                        'choice': 'basic-test-choice',
                        'op': 'basic-test-op',
                        'value': 'my-value',
                    },
                ],
            })

        self.assertEqual(condition_set.mode, ConditionSet.MODE_ANY)
        self.assertEqual(len(condition_set.conditions), 1)
        self.assertEqual(condition_set.conditions[0].choice.choice_id,
                         'basic-test-choice')

    def test_deserialize_with_choice_kwargs(self):
        """Testing ConditionSet.deserialize with choice_kwargs"""
        choices = ConditionChoices([BasicTestChoice])

        condition_set = ConditionSet.deserialize(
            choices,
            {
                'mode': 'any',
                'conditions': [
                    {
                        'choice': 'basic-test-choice',
                        'op': 'basic-test-op',
                        'value': 'my-value',
                    },
                ],
            },
            choice_kwargs={
                'abc': 123,
            })

        self.assertEqual(condition_set.mode, ConditionSet.MODE_ANY)
        self.assertEqual(len(condition_set.conditions), 1)

        choice = condition_set.conditions[0].choice
        self.assertEqual(choice.choice_id, 'basic-test-choice')
        self.assertEqual(choice.extra_state, {'abc': 123})

    def test_deserialize_with_invalid_mode(self):
        """Testing ConditionSet.deserialize with invalid mode"""
        choices = ConditionChoices([BasicTestChoice])

        with self.assertRaises(InvalidConditionModeError):
            ConditionSet.deserialize(
                choices,
                {
                    'mode': 'invalid',
                    'conditions': [
                        {
                            'choice': 'basic-test-choice',
                            'op': 'basic-test-op',
                            'value': 'my-value',
                        },
                    ],
                })

    def test_matches_with_always_mode(self):
        """Testing ConditionSet.matches with "always" mode"""
        condition_set = ConditionSet(ConditionSet.MODE_ALWAYS, [])
        self.assertTrue(condition_set.matches(value='abc123'))

    def test_matches_with_all_mode_and_match(self):
        """Testing ConditionSet.matches with "all" mode and match"""
        choice = EqualsTestChoice()

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(choice, choice.get_operator('equals-test-op'), 'abc123'),
            Condition(choice, choice.get_operator('equals-test-op'), 'abc123'),
        ])

        self.assertTrue(condition_set.matches(value='abc123'))

    def test_matches_with_all_mode_and_no_match(self):
        """Testing ConditionSet.matches with "all" mode and no match"""
        choice = EqualsTestChoice()

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(choice, choice.get_operator('equals-test-op'), 'abc123'),
            Condition(choice, choice.get_operator('equals-test-op'), 'def123'),
        ])

        self.assertFalse(condition_set.matches(value='abc123'))

    def test_matches_with_any_mode_and_match(self):
        """Testing ConditionSet.matches with "any" mode and match"""
        choice = EqualsTestChoice()

        condition_set = ConditionSet(ConditionSet.MODE_ANY, [
            Condition(choice, choice.get_operator('equals-test-op'), 'abc123'),
            Condition(choice, choice.get_operator('equals-test-op'), 'def123'),
        ])

        self.assertTrue(condition_set.matches(value='abc123'))

    def test_matches_with_any_mode_and_no_match(self):
        """Testing ConditionSet.matches with "any" mode and no match"""
        choice = EqualsTestChoice()

        condition_set = ConditionSet(ConditionSet.MODE_ANY, [
            Condition(choice, choice.get_operator('equals-test-op'), 'abc123'),
            Condition(choice, choice.get_operator('equals-test-op'), 'def123'),
        ])

        self.assertFalse(condition_set.matches(value='foo'))

    def test_matches_with_custom_value_kwargs(self):
        """Testing ConditionSet.matches with custom value keyword arguments"""
        class CustomEqualsChoice(EqualsTestChoice):
            value_kwarg = 'my_value'

        choice = CustomEqualsChoice()

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(choice, choice.get_operator('equals-test-op'),
                      'abc123'),
        ])

        self.assertTrue(condition_set.matches(my_value='abc123'))
        self.assertFalse(condition_set.matches(value='abc123'))

    def test_matches_with_all_mode_and_custom_value_kwargs_multiple(self):
        """Testing ConditionSet.matches with "all" mode and multiple custom
        value keyword arguments across multiple choices
        """
        class CustomEqualsChoice1(EqualsTestChoice):
            value_kwarg = 'my_value1'

        class CustomEqualsChoice2(EqualsTestChoice):
            value_kwarg = 'my_value2'

        choice1 = CustomEqualsChoice1()
        choice2 = CustomEqualsChoice2()

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(choice1, choice1.get_operator('equals-test-op'),
                      'abc123'),
            Condition(choice2, choice2.get_operator('equals-test-op'),
                      'def456'),
        ])

        self.assertTrue(condition_set.matches(my_value1='abc123',
                                              my_value2='def456'))
        self.assertFalse(condition_set.matches(my_value1='abc123'))
        self.assertFalse(condition_set.matches(my_value2='def456'))
        self.assertFalse(condition_set.matches(my_value1='abc123',
                                               my_value2='xxx'))

    def test_matches_with_any_mode_and_custom_value_kwargs_multiple(self):
        """Testing ConditionSet.matches with "any" mode and multiple custom
        value keyword arguments across multiple choices
        """
        class CustomEqualsChoice1(EqualsTestChoice):
            value_kwarg = 'my_value1'

        class CustomEqualsChoice2(EqualsTestChoice):
            value_kwarg = 'my_value2'

        choice1 = CustomEqualsChoice1()
        choice2 = CustomEqualsChoice2()

        condition_set = ConditionSet(ConditionSet.MODE_ANY, [
            Condition(choice1, choice1.get_operator('equals-test-op'),
                      'abc123'),
            Condition(choice2, choice2.get_operator('equals-test-op'),
                      'def456'),
        ])

        self.assertTrue(condition_set.matches(my_value1='abc123',
                                              my_value2='def456'))
        self.assertTrue(condition_set.matches(my_value1='abc123'))
        self.assertTrue(condition_set.matches(my_value2='def456'))
        self.assertTrue(condition_set.matches(my_value1='abc123',
                                              my_value2='xxx'))
        self.assertFalse(condition_set.matches(my_value1='xxx',
                                               my_value2='xxx'))

    def test_serialize(self):
        """Testing ConditionSet.serialize"""
        basic_choice = BasicTestChoice()
        equals_choice = EqualsTestChoice()

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(basic_choice,
                      basic_choice.get_operator('basic-test-op'),
                      'abc123'),
            Condition(equals_choice,
                      equals_choice.get_operator('equals-test-op'),
                      'def123'),
        ])

        result = condition_set.serialize()

        self.assertEqual(
            result,
            {
                'mode': 'all',
                'conditions': [
                    {
                        'choice': 'basic-test-choice',
                        'op': 'basic-test-op',
                        'value': 'abc123',
                    },
                    {
                        'choice': 'equals-test-choice',
                        'op': 'equals-test-op',
                        'value': 'def123',
                    },
                ],
            })



"""Error classes for features."""

from djblets.registries.errors import AlreadyRegisteredError, ItemLookupError


class FeatureConflictError(AlreadyRegisteredError):
    """A feature was not found."""


class FeatureNotFoundError(ItemLookupError):
    """A feature was not found."""



"""E-mail formatting template tags."""

import re

from django import template
from django.template.loader import render_to_string
from django.utils.safestring import mark_safe

from djblets.util.decorators import blocktag


register = template.Library()


@register.simple_tag(takes_context=True)
def quoted_email(context, template_name):
    """
    Renders a specified template as a quoted reply, using the current context.
    """
    return quote_text(render_to_string(template_name, context.flatten()))


@register.tag
@blocktag
def condense(context, nodelist, max_newlines=3):
    """Condenses a block of text.

    This will ensure that there are never more than the given number of
    consecutive newlines. It's particularly useful when formatting plain text
    output, to avoid issues with template tags adding unwanted newlines.
    """
    text = nodelist.render(context).strip()
    text = re.sub(r'\n{%d,}' % (max_newlines + 1), '\n' * max_newlines, text)
    return text


@register.filter
def quote_text(text, level=1):
    """
    Quotes a block of text the specified number of times.
    """
    lines = text.split("\n")
    quoted = ""

    for line in lines:
        quoted += "%s%s\n" % ("> " * level, line)

    return mark_safe(quoted.rstrip())



"""Mixins for integrating a web API resource with a form."""

from django.core.exceptions import ValidationError
from django.forms.forms import NON_FIELD_ERRORS
from django.forms.models import model_to_dict
from django.utils.encoding import force_str

from djblets.webapi.errors import INVALID_FORM_DATA


class UpdateFormMixin(object):
    """A mixin for providing the ability to create and update using a form.

    A WebAPIResource class using this mixin must set the :py:attr:`form_class`
    attribute to a :py:class:`~django.forms.ModelForm` instance that
    corresponds to the model being updated.

    Classes using this mixin can provide methods of the form
    :samp:`parse_{field_name}_field` to do parsing of form data before it is
    passed to the form. Parser methods should be of the form:

    .. code-block:: python

      def parse_some_field(self, value, request, **kwargs):
          return some_value

    These methods may return either a single value or a list of values (in
    the case where the corresponding field expects a list of values, such as a
    :py:class:`~django.forms.ModelMultipleChoiceField`). They may also raise a
    :py:class:`~django.core.exceptions.ValidationError`, though it's up to the
    caller of :py:meth:`create_form` to catch this and return any suitable
    errors.

    Most implementations will want to call :py:meth:`handle_form_request` in
    their POST/PUT handlers, and override behavior with the parsing methods.
    Some may also want to override :py:meth:`save_form`,
    :py:meth:`build_form_success_response`, or
    :py:meth:`build_form_error_response` to customize behavior.
    """

    #: The form class for updating models.
    #:
    #: This should be a subclass of :py:class:`django.forms.ModelForm`.
    form_class = None

    @property
    def add_form_class(self):
        """The form class for creating new models.

        This should be a subclass of :py:class:`django.forms.ModelForm`. It
        defaults to :py:attr:`form_class`.
        """
        return self.form_class

    def handle_form_request(self, request, data=None, files=None,
                            instance=None, form_kwargs=None,
                            save_kwargs=None, **kwargs):
        """Handle an HTTP request for creating or updating through a form.

        This can be called directly from a resource's
        :py:meth:`~djblets.webapi.resources.base.WebAPIResource.create` or
        :py:meth:`~djblets.webapi.resources.base.WebAPIResource.update` method
        to parse request data, create the form, and handle errors or the
        saving of the form.

        Version Added:
            1.0.9

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            data (dict, optional):
                The data to pass to :py:meth:`create_form`.

            files (dict, optional):
                Files to pass to the form.

            instance (django.db.models.Model, optional):
                An existing instance to update, if performing an HTTP PUT
                request.

            form_kwargs (dict, optional):
                Keyword arguments to pass to the form's constructor.

            save_kwargs (dict, optional):
                Keyword arguments to pass to the form's
                :py:meth:`ModelForm.save() <django.forms.ModelForm.save>`
                method.

            **kwargs (dict):
                Keyword arguments to pass to
                :py:meth:`build_form_error_response`,
                :py:meth:`build_form_success_response`,
                :py:meth:`create_form`,
                :py:meth:`save_form`, and any field parsing methods.

        Returns:
            tuple or django.http.HttpResponse:
            The response to send back to the client.
        """
        is_created = instance is None

        try:
            form = self.create_form(instance=instance,
                                    data=data,
                                    files=files,
                                    request=request,
                                    form_kwargs=form_kwargs,
                                    **kwargs)
        except ValidationError as e:
            return self.build_form_error_response(errors=e,
                                                  instance=instance,
                                                  **kwargs)

        if not form.is_valid():
            return self.build_form_error_response(form=form,
                                                  instance=instance,
                                                  **kwargs)

        instance = self.save_form(form=form,
                                  save_kwargs=save_kwargs,
                                  **kwargs)

        return self.build_form_success_response(form=form,
                                                instance=instance,
                                                is_created=is_created,
                                                **kwargs)

    def create_form(self, data, request, files=None, instance=None,
                    form_kwargs=None, **kwargs):
        """Create a new form and pre-fill it with data.

        Version Changed:
            1.0.9:
            The initial values for form fields are now automatically provided
            in the form data, if not otherwise overridden, making it easier to
            construct forms.

            Along with this, a :py:class:`~django.forms.ModelForm`'s ``fields``
            and ``exclude`` lists are now factored in when populating the formw
            with an instance's data.

        Args:
            data (dict):
                The request data to pass to the form.

            request (django.http.HttpRequest):
                The HTTP request.

            files (dict):
                Files to pass to the form.

            instance (django.db.models.Model, optional):
                The instance model, if it exists. If this is not ``None``,
                fields that appear in the form class's ``fields`` attribute
                that do not appear in the ``data`` dict as keys will be copied
                from the instance.

            form_kwargs (dict, optional):
                Additional keyword arguments to provide to the form's
                constructor.

            **kwargs (dict):
                Additional arguments. These will be passed to the resource's
                parser methods.

                This contains anything passed as keyword arguments to
                :py:meth:`handle_form_request`.

        Returns:
            django.forms.ModelForm:
            The form with data filled.

        Raises:
            django.core.exceptions.ValidationError:
                A field failed validation. This is allowed to be raised by
                any ``parse_*`` methods defined on the resource.
        """
        assert self.form_class, ('%s must define a form_class attribute.'
                                 % self.__class__.__name__)

        if instance is not None:
            form_cls = self.form_class
        else:
            form_cls = self.add_form_class

        form_data = self._get_initial_form_data(form_cls)

        if instance is not None:
            meta = form_cls._meta
            form_data.update(model_to_dict(instance=instance,
                                           fields=meta.fields,
                                           exclude=meta.exclude))

        form_data.update(self._parse_form_data(data, request, **kwargs))

        # Dynamically provide the arguments we want to the form, so that
        # any form classes lacking an argument (files, for instance) won't
        # result in a crash so long as the equivalent argument is not provided.
        if form_kwargs is None:
            form_kwargs = {}
        else:
            form_kwargs = form_kwargs.copy()

        if form_data is not None:
            form_kwargs['data'] = form_data

        if files is not None:
            form_kwargs['files'] = files

        if instance is not None:
            form_kwargs['instance'] = instance

        return form_cls(**form_kwargs)

    def save_form(self, form, save_kwargs=None, **kwargs):
        """Save and return the results of the form.

        This is a simple wrapper around calling :py:meth:`ModelForm.save()
        <django.forms.ModelForm.save>`. It can be overridden by subclasses
        that need to perform additional operations on the instance or form.

        Version Added:
            1.0.9

        Args:
            form (django.forms.Form):
                The form to save.

            save_kwargs (dict):
                Any keyword arguments to pass when saving the form.

            **kwargs (dict):
                Additional keyword arguments passed by the caller. This
                contains anything passed as keyword arguments to
                :py:meth:`handle_form_request`.

        Returns:
            django.db.models.Model:
            The saved model instance.
        """
        return form.save(**(save_kwargs or {}))

    def build_form_errors(self, form=None, errors=None, **kwargs):
        """Return a dictionary of field errors for use in a response payload.

        This will convert each error to a string, resulting in a dictionary
        mapping field names to lists of errors. This can be safely returned in
        any API payload.

        Version Added:
            1.0.9

        Args:
            form (django.forms.Form, optional):
                The form containing errors. This may be ``None`` if handling
                a :py:class:`~django.core.exceptions.ValidationError` from
                field parsing.

            errors (django.core.exceptions.ValidationError, optional):
                An explicit validation error to use for the payload. This will
                be used if field parsing fails.

            **kwargs (dict):
                Additional keyword arguments passed by the caller. This
                contains anything passed as keyword arguments to
                :py:meth:`handle_form_request`.

        Returns:
            dict:
            The dictionary of errors. Each key is a field name and each
            value is a list of error strings.
        """
        assert (form is None) != (errors is None), \
            'Only one of form or errors can be provided.'

        if form is not None:
            return {
                field_name: [force_str(e) for e in field_errors]
                for field_name, field_errors in form.errors.items()
            }
        else:
            if hasattr(errors, 'error_dict'):
                return errors.message_dict
            else:
                return {
                    NON_FIELD_ERRORS: errors.messages,
                }

    def build_form_success_response(self, form, instance, is_created,
                                    **kwargs):
        """Return a success response for a saved instance.

        Version Added:
            1.0.9

        Args:
            form (django.forms.Form):
                The form that saved the instance.

            instance (django.db.models.Model):
                The saved instance.

            is_created (bool):
                Whether the instance was created.

            **kwargs (dict):
                Additional keyword arguments passed by the caller. This
                contains anything passed as keyword arguments to
                :py:meth:`handle_form_request`.

        Returns:
            tuple or django.http.HttpResponse:
            The success response to return from the API handler.
        """
        if is_created:
            code = 201
        else:
            code = 200

        return code, {
            self.item_result_key: instance,
        }

    def build_form_error_response(self, form=None, errors=None, instance=None,
                                  **kwargs):
        """Return an error response containing errors for form fields.

        Version Added:
            1.0.9

        Args:
            form (django.forms.Form, optional):
                The form containing errors. This may be ``None`` if handling
                a :py:class:`~django.core.exceptions.ValidationError` from
                field parsing.

            errors (django.core.exceptions.ValidationError, optional):
                An explicit validation error to use for the payload. This will
                be used if field parsing fails.

            instance (django.db.models.Model, optional):
                The existing instance, if any.

            **kwargs (dict):
                Additional keyword arguments passed by the caller. This
                contains anything passed as ``response_kwargs`` to
                :py:meth:`handle_form_request`.

        Returns:
            tuple or django.http.HttpResponse:
            The error response to return from the API handler.
        """
        return INVALID_FORM_DATA, {
            'fields': self.build_form_errors(form=form,
                                             errors=errors),
        }

    def _get_initial_form_data(self, form_cls):
        """Return the initial data for a form.

        This will return the initial data from all fields in a form,
        converting each into a format suitable for passing as bound data
        for a form.

        Args:
            form_cls (type):
                The form's class.

        Returns:
            dict:
            The initial data for the fields in the form.
        """
        initial = {}

        for field_name, field in form_cls.base_fields.items():
            data = field.initial

            if callable(data):
                data = data()

            initial[field_name] = data

        return initial

    def _parse_form_data(self, form_data, request, **kwargs):
        """Parse the form data.

        Args:
            form_data (dict):
                The request data.

            request (django.http.HttpRequest):
                The HTTP request.

            **kwargs (dict):
                Additional arguments to pass to parser methods.

        Returns:
            dict:
            A mapping of field names to parsed values.

        Raises:
            django.core.exceptions.ValidationError:
                A field failed validation containing errors for any fields
                that failed to parse.
        """
        parsed_data = form_data.copy()
        errors = {}

        for field, value in form_data.items():
            parser = getattr(self, 'parse_%s_field' % field, None)

            if parser is not None:
                try:
                    parsed_data[field] = parser(value, request=request,
                                                **kwargs)
                except ValidationError as e:
                    errors[field] = e.messages
            else:
                parsed_data[field] = value

        if errors:
            raise ValidationError(errors)

        return parsed_data



"""Unit tests for LegacySHA1TokenGenerator.

Version Added:
    3.0
"""

from django.contrib.auth.models import User

from djblets.secrets.token_generators.legacy_sha1 import \
    LegacySHA1TokenGenerator
from djblets.testing.testcases import TestCase


class LegacySHA1TokenGeneratorTests(TestCase):
    """Unit tests for LegacySHA1TokenGenerator.

    Version Added:
        3.0
    """

    def setUp(self):
        super().setUp()

        self.user = User.objects.create(username='test-user')
        self.token_generator = LegacySHA1TokenGenerator()

    def test_create_token(self):
        """Testing LegacySHA1TokenGenerator.create_token"""
        token = self.token_generator.create_token(
            token_info={
                'user': self.user,
                'attempt': 0,
            })

        self.assertIsNotNone(token)

    def test_create_token_with_missing_required_key(self):
        """Testing LegacySHA1TokenGenerator.create_token with a missing
        required key in token info
        """
        error_message = 'The token_info dictionary must contain a user key.'

        with self.assertRaisesMessage(KeyError, error_message):
            self.token_generator.create_token(
                token_info={
                    'attempt': 0,
                })

    def test_validate_token_with_valid_token(self):
        """Testing LegacySHA1TokenGenerator.validate_token with valid token"""
        token = 'ead9190eadc2f8ccc051f813be28ea1f24ba3810'

        self.assertTrue(self.token_generator.validate_token(token))

    def test_validate_token_with_invalid_token_length(self):
        """Testing LegacySHA1TokenGenerator.validate_token with a token that
        has an invalid length
        """
        token = 'abcd1'

        self.assertFalse(self.token_generator.validate_token(token))

    def test_validate_token_with_invalid_token_chars(self):
        """Testing LegacySHA1TokenGenerator.validate_token with a token that
        has an invalid character in it
        """
        token = '!ad9190eadc2f8ccc051f813be28ea1f24ba3810'

        self.assertFalse(self.token_generator.validate_token(token))



import threading

from django.urls import (URLResolver,
                         clear_url_caches,
                         get_resolver)
from django.urls.resolvers import RegexPattern


class DynamicURLResolver(URLResolver):
    """A URL resolver that allows for dynamically altering URL patterns.

    A standard URLResolver expects that a list of URL patterns will be set
    once and never again change. In most applications, this is a good
    assumption. However, some that are more specialized may need to be able
    to swap in URL patterns dynamically. For example, those that can plug in
    third-party extensions.

    DynamicURLResolver makes it easy to add and remove URL patterns. Any
    time the list of URL patterns changes, they'll be immediately available
    for all URL resolution and reversing.

    The usage is very simple::

        dynamic_patterns = DynamicURLResolver()
        urlpatterns = [dynamic_patterns]

        dynamic_patterns.add_patterns([
            url(...),
            url(...),
        ])

    DynamicURLResolver will handle managing all the lookup caches to ensure
    that there won't be any stale entries affecting any dynamic URL patterns.
    """

    def __init__(self, regex=r'', app_name=None, namespace=None):
        """Initialize the resolver.

        Args:
            regex (unicode, optional):
                The relative pattern for the root of any child URLs.

            app_name (unicode, optional):
                The application namespace for any child URLs.

            namespace (unicode, optional):
                The instance namespace for any child URLs.
        """
        super(DynamicURLResolver, self).__init__(pattern=RegexPattern(regex),
                                                 urlconf_name=[],
                                                 app_name=app_name,
                                                 namespace=namespace)
        self._resolver_chain = None
        self._lock = threading.Lock()

    @property
    def url_patterns(self):
        """Returns the current list of URL patterns.

        This is a simplified version of URLResolver.url_patterns that
        simply returns the preset list of patterns. Unlike the original
        function, we don't care if the list is empty.
        """
        # Internally, urlconf_module represents whatever we're accessing
        # for the list of URLs. It can be a list, or it can be something
        # with a 'urlpatterns' property (intended for a urls.py). However,
        # we force this to be a list in the constructor (as urlconf_name,
        # which gets stored as urlconf_module), so we know we can just
        # return it as-is.
        return self.urlconf_module

    def add_patterns(self, patterns):
        """Adds a list of URL patterns.

        The patterns will be made immediately available for use for any
        lookups or reversing.
        """
        self.url_patterns.extend(patterns)
        self._repopulate_caches()

    def remove_patterns(self, patterns):
        """Removes a list of URL patterns.

        These patterns will no longer be able to be looked up or reversed.
        """
        for pattern in patterns:
            try:
                self.url_patterns.remove(pattern)
            except ValueError:
                # This may have already been removed. Ignore the error.
                pass

        self._repopulate_caches()

    def _repopulate_caches(self):
        """Repopulates the internal resolver caches.

        This will force all the resolvers in the chain to repopulate,
        replacing the caches, in order to ensure that the next lookup or
        reverse will result in a lookup in this resolver. By default, every
        URLResolver in Django will cache all results from its children.

        We take special care to only repopulate the caches of the resolvers in
        our parent chain.
        """
        with self._lock:
            for resolver in self.resolver_chain:
                # Clear the caches for the URL resolvers. This will ensure
                # that the next request on this resolver will re-populate it.
                #
                # This is known to work at least through Django 1.9.
                resolver._reverse_dict = {}
                resolver._namespace_dict = {}
                resolver._app_dict = {}
                resolver._callback_strs = set()
                resolver._populated = False

            clear_url_caches()

    @property
    def resolver_chain(self):
        """Returns every URLResolver between here and the root.

        The list of resolvers is cached in order to prevent having to locate
        the resolvers more than once.
        """
        if self._resolver_chain is None:
            self._resolver_chain = \
                self._find_resolver_chain(get_resolver(None))

        return self._resolver_chain

    def _find_resolver_chain(self, resolver):
        if resolver == self:
            return [resolver]

        for url_pattern in resolver.url_patterns:
            if isinstance(url_pattern, URLResolver):
                resolvers = self._find_resolver_chain(url_pattern)

                if resolvers:
                    resolvers.append(resolver)
                    return resolvers

        return []



"""Decorative template tags."""

from django import template
from django.template.loader import render_to_string

from djblets.util.decorators import blocktag


register = template.Library()


@register.tag
@blocktag
def box(context, nodelist, classname=None):
    """
    Displays a box container around content, with an optional class name.
    """
    return render_to_string('deco/box.html', {
        'classname': classname or "",
        'content': nodelist.render(context)
    })


@register.tag
@blocktag
def errorbox(context, nodelist, box_id=None):
    """
    Displays an error box around content, with an optional ID.
    """
    return render_to_string('deco/errorbox.html', {
        'box_id': box_id or "",
        'content': nodelist.render(context)
    })



"""Unit tests for djblets.db.fields.modification_timestamp_field."""

from datetime import datetime, timedelta

from django.db import models
from django.utils import timezone
from kgb import SpyAgency

from djblets.db.fields.modification_timestamp_field import \
    ModificationTimestampField
from djblets.testing.testcases import TestCase, TestModelsLoaderMixin


class ModificationTimestampFieldTestModel(models.Model):
    field = ModificationTimestampField()


class MultitipleModificationTimestampFieldTestModel(models.Model):
    timestamp1 = ModificationTimestampField('timestamp1')
    timestamp2 = ModificationTimestampField('timestamp2')


class ModificationTimestampFieldTests(SpyAgency, TestModelsLoaderMixin,
                                      TestCase):
    """Unit tests for ModificationTimestampField."""

    tests_app = 'djblets.db.tests'

    def test_pre_save_with_new_instance_with_auto_value(self):
        """Testing ModificationTimestampField.pre_save with new model instance
        and auto-added value
        """
        obj = ModificationTimestampFieldTestModel()
        self.assertIsNone(obj.field)

        obj.save()
        self.assertIsNotNone(obj.field)

    def test_pre_save_with_new_instance_with_custom_value(self):
        """Testing ModificationTimestampField.pre_save with new model instance
        and custom value
        """
        timestamp = datetime(2018, 3, 6, 23, 31, 30, tzinfo=timezone.utc)

        obj = ModificationTimestampFieldTestModel(field=timestamp)
        self.assertIsNotNone(obj.field)

        obj.save()
        self.assertIsNotNone(obj.field)
        self.assertEqual(obj.field, timestamp)

    def test_pre_save_with_existing_instance(self):
        """Testing ModificationTimestampField.pre_save with existing model
        instance
        """
        timestamp = datetime(2018, 3, 6, 23, 31, 30, tzinfo=timezone.utc)

        obj = ModificationTimestampFieldTestModel.objects.create(
            field=timestamp)
        self.assertEqual(obj.field, timestamp)

        obj.save()
        self.assertNotEqual(obj.field, timestamp)
        self.assertTrue(obj.field > timestamp)

    def test_get_internal_type(self):
        """Testing ModificationTimestampField.get_internal_type"""
        obj = ModificationTimestampFieldTestModel()
        self.assertEqual(obj._meta.get_field('field').get_internal_type(),
                         'DateTimeField')

    def test_save_updates_value_with_tz(self):
        """Testing ModificationTimestampField updates the value on save when
        USE_TZ=True
        """
        with self.settings(USE_TZ=True):
            original_timestamp = timestamp = timezone.now()
            self.spy_on(timezone.now, call_fake=lambda: timestamp)

            instance = ModificationTimestampFieldTestModel.objects.create()
            self.assertEqual(instance.field, timestamp)

            timestamp += timedelta(hours=1)
            instance.save()
            self.assertEqual(instance.field, timestamp)
            self.assertNotEqual(instance.field, original_timestamp)

    def test_save_updates_value_without_tz(self):
        """Testing ModificationTimestampField udpates the value on save when
        USE_TZ=False
        """
        with self.settings(USE_TZ=False):
            original_timestamp = timestamp = timezone.now()
            self.spy_on(timezone.now, call_fake=lambda: timestamp)

            instance = ModificationTimestampFieldTestModel.objects.create()
            self.assertEqual(instance.field, timestamp)

            timestamp += timedelta(hours=1)
            instance.save()
            self.assertEqual(instance.field, timestamp)
            self.assertNotEqual(instance.field, original_timestamp)

    def test_save_doesnt_overwrite_modified_value_with_tz(self):
        """Testing ModificationTimestampField doesn't overwite manual setting
        of the field when USE_TZ=False
        """
        with self.settings(USE_TZ=True):
            timestamp = timezone.now()
            self.spy_on(timezone.now, call_fake=lambda: timestamp)

            instance = ModificationTimestampFieldTestModel.objects.create()
            instance.field += timedelta(hours=1)
            instance.save()

            self.assertNotEqual(instance.field, timestamp)
            self.assertEqual(instance.field,
                             timestamp + timedelta(hours=1))

    def test_save_doesnt_overwite_modified_value_without_tz(self):
        """Testing ModificationTimestampField doesn't overwite manual setting
        of the field when USE_TZ=False
        """
        with self.settings(USE_TZ=False):
            timestamp = timezone.now()
            self.spy_on(timezone.now, call_fake=lambda: timestamp)

            instance = ModificationTimestampFieldTestModel.objects.create()
            instance.field += timedelta(hours=1)
            instance.save()

            self.assertNotEqual(instance.field, timestamp)
            self.assertEqual(instance.field,
                             timestamp + timedelta(hours=1))

    def test_save_multiple_with_tz(self):
        """Testing multiple ModificationTimestampField save correctly when
        USE_TZ=True
        """
        with self.settings(USE_TZ=True):
            timestamp = timezone.now()
            self.spy_on(timezone.now, call_fake=lambda: timestamp)

            manual_value = timestamp + timedelta(hours=4)
            timestamp += timedelta(hours=1)

            instance = MultitipleModificationTimestampFieldTestModel()
            instance.timestamp1 = manual_value
            instance.save()

            self.assertEqual(instance.timestamp1, manual_value)
            self.assertEqual(instance.timestamp2, timestamp)

            timestamp += timedelta(hours=1)

            instance.timestamp2 = manual_value
            instance.save()

            self.assertEqual(instance.timestamp1, timestamp)
            self.assertEqual(instance.timestamp2, manual_value)

    def test_save_multiple_without_tz(self):
        """Testing multiple ModificationTimestampField save correctly when
        USE_TZ=False
        """
        with self.settings(USE_TZ=False):
            timestamp = timezone.now()
            self.spy_on(timezone.now, call_fake=lambda: timestamp)

            manual_value = timestamp + timedelta(hours=4)
            timestamp += timedelta(hours=1)

            instance = MultitipleModificationTimestampFieldTestModel()
            instance.timestamp1 = manual_value
            instance.save()

            self.assertEqual(instance.timestamp1, manual_value)
            self.assertEqual(instance.timestamp2, timestamp)

            timestamp += timedelta(hours=1)

            instance.timestamp2 = manual_value
            instance.save()

            self.assertEqual(instance.timestamp1, timestamp)
            self.assertEqual(instance.timestamp2, manual_value)



from contextlib import contextmanager

from django.contrib.sites.models import Site

from djblets.testing.testcases import TestCase
from djblets.webapi.resources.base import WebAPIResource
from djblets.webapi.resources.registry import (get_resource_for_object,
                                               register_resource_for_model,
                                               unregister_resource_for_model)


@contextmanager
def register_resource_for_model_temp(model, resource):
    """A context manager to temporarily register a resource for a model."""
    register_resource_for_model(model, resource)

    try:
        yield
    finally:
        unregister_resource_for_model(model)


class ResourceRegistryTests(TestCase):
    """Tests for the resource registry."""

    def test_get_resource_for_object_only_fields(self):
        """Testing get_resource_for_model when model is deferred"""
        class TestResource(WebAPIResource):
            pass

        resource = TestResource()

        with register_resource_for_model_temp(Site, resource):
            config = Site.objects.only('domain').get()

            self.assertEqual(get_resource_for_object(config),
                             resource)



"""Registry for managing feature registrations."""

from django.utils.translation import gettext_lazy as _

from djblets.features.errors import FeatureConflictError, FeatureNotFoundError
from djblets.registries.registry import (ALREADY_REGISTERED,
                                         ATTRIBUTE_REGISTERED, DEFAULT_ERRORS,
                                         RegistrationError, Registry,
                                         UNREGISTER)


FEATURE_DEFAULT_ERRORS = DEFAULT_ERRORS.copy()
FEATURE_DEFAULT_ERRORS.update({
    ALREADY_REGISTERED: _(
        'Could not register feature %(item)s: This feature is already '
        'registered or its ID conflicts with another feature.',
    ),
    ATTRIBUTE_REGISTERED: _(
        'Could not register feature %(item)s: Another feature (%(duplicate)s) '
        'is already registered with the same feature ID.',
    ),
    UNREGISTER: _(
        'Could not unregister feature %(item)s: This feature was not yet '
        'registered.',
    ),
})


_registry = None


class FeaturesRegistry(Registry):
    """A registry for instantiated features.

    This manages all instances of :py:class:`~djblets.features.feature.Feature`
    subclasses that the product has created, providing easy access to features
    for checking purposes.
    """

    lookup_attrs = ('feature_id',)
    default_errors = FEATURE_DEFAULT_ERRORS
    already_registered_error_class = FeatureConflictError
    lookup_error_class = FeatureNotFoundError

    def register(self, feature):
        """Register a feature instance.

        The feature's :py:meth:`~djblets.features.feature.Feature.initialize`
        method will be called once registered.

        Args:
            feature (djblets.features.feature.Feature):
                The feature to register.

        Raises:
            djblets.features.errors.FeatureConflictError:
                The feature's ID conflicts with another feature class.

            djblets.registries.errors.RegistrationError:
                The feature ID wasn't set on the class.
        """
        if not feature.feature_id:
            raise RegistrationError("The feature class's ID must be set.")

        feature.initialize()

        super(FeaturesRegistry, self).register(feature)

    def unregister(self, feature):
        """Unregister a feature instance.

        The feature's :py:meth:`~djblets.features.feature.Feature.shutdown`
        method will be called once unregistered.

        Args:
            feature (djblets.features.feature.Feature):
                The feature to unregister.

        Raises:
            djblets.features.errors.FeatureNotFoundError:
                Raised if the feature was not already registered.
        """
        super(FeaturesRegistry, self).unregister(feature)

        feature.shutdown()

    def get_feature(self, feature_id):
        """Return the feature instance with the given ID.

        Args:
            feature_id (unicode):
                The ID of the feature to return.

        Returns:
            djblets.features.feature.Feature:
            The feature instance matching the ID or ``None`` if not found.
        """
        try:
            return self.get('feature_id', feature_id)
        except FeatureNotFoundError:
            return None


def get_features_registry():
    """Return the global features registry.

    The first time this is called, a :py:class:`FeaturesRegistry` will be
    instantiated and cached for future calls.

    Returns:
        FeaturesRegistry:
        The features registry.
    """
    global _registry

    if _registry is None:
        _registry = FeaturesRegistry()

    return _registry



"""Compatibility for collections ABCs removed in Python 3.10.

Note that this module is not considered API-stable. Changes may be made that
remove or alter functionality in the future. Please use at your own risk.

Version Added:
    2.2.4
"""

import sys


if sys.version_info >= (3, 10):
    # Python 3.10+ removed a bunch of aliases from the collections module.
    # Unfortunately, Django 1.11 and nose still rely on these.
    try:
        from collections import Callable
    except ImportError:
        import collections
        import collections.abc

        collections.Callable = collections.abc.Callable
        collections.Iterable = collections.abc.Iterable
        collections.Iterator = collections.abc.Iterator
        collections.MutableMapping = collections.abc.MutableMapping
        collections.Mapping = collections.abc.Mapping
        collections.Sequence = collections.abc.Sequence



from djblets.extensions.hooks import ExtensionHookPoint
from djblets.integrations.hooks import BaseIntegrationHook
from djblets.integrations.integration import Integration
from djblets.integrations.manager import IntegrationManager
from djblets.integrations.tests.models import IntegrationConfig
from djblets.integrations.tests.testcases import IntegrationsTestCase


class DummyIntegrationHook(BaseIntegrationHook, metaclass=ExtensionHookPoint):
    def __init__(self, manager, *args, **kwargs):
        self.manager = manager

        super(DummyIntegrationHook, self).__init__(*args, **kwargs)

    def get_integration_manager(self):
        return self.manager


class DummyIntegration(Integration):
    def initialize(self):
        pass


class IntegrationHookTests(IntegrationsTestCase):
    """Unit tests for djblets.integrations.hooks.BaseIntegrationHook."""

    def setUp(self):
        super(IntegrationHookTests, self).setUp()

        self.manager = IntegrationManager(IntegrationConfig)

        # This will make this class a hook owner.
        self.hooks = set()

        self.hook = DummyIntegrationHook(self.manager, self, DummyIntegration)

    def test_initialize(self):
        """Testing BaseIntegrationHook.initialize"""
        self.assertIn(DummyIntegration.integration_id,
                      self.manager._integration_classes)

    def test_shutdown(self):
        """Testing BaseIntegrationHook.initialize"""
        self.hook.disable_hook()

        self.assertNotIn(DummyIntegration.integration_id,
                         self.manager._integration_classes)



"""Token generator for legacy SHA1 tokens.

Version Added:
    3.0
"""

import hashlib
import re
from typing import Any, Dict

from django.conf import settings
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from djblets.secrets.token_generators.base import BaseTokenGenerator


class LegacySHA1TokenGenerator(BaseTokenGenerator):
    """Token generator for legacy SHA1 tokens.

    The support for SHA1 tokens will soon be deprecated in favour of
    :py:class:`~djblets.secrets.token_generators.vendor_checksum.
    VendorChecksumTokenGenerator` tokens.

    Version Added:
        3.0
    """

    token_generator_id = 'legacy_sha1'

    def create_token(
        self,
        token_info: Dict[str, Any] = {},
        **kwargs,
    ) -> str:
        """Create a SHA1 token.

        Args:
            token_info (dict):
                A dictionary that must contain the following information
                about the token:

                Keys:
                    attempt (int):
                        The number of attempts that have been made to create
                        a token. This is used in the raw token string.

                    user (django.contrib.auth.models.User):
                        The user that the token belongs to.

            **kwargs (dict, unused):
                Additional keyword arguments.

        Returns:
            str:
            The token value.

        Raises:
            KeyError:
                The ``token_info`` dictionary is missing a required key.
        """
        try:
            user = token_info['user']
            attempt = token_info['attempt']
        except KeyError as e:
            raise KeyError(
                _('The token_info dictionary must contain a %s key.')
                % e.args[0])

        prefix = '%s%s%s' % (settings.SECRET_KEY, str(user.pk), user.password)

        raw_token = '%s%s%s' % (prefix, str(attempt),
                                timezone.now().isoformat())
        sha = hashlib.sha1(raw_token.encode('utf-8'))
        token = sha.hexdigest()

        return token

    def validate_token(
        self,
        token: str,
        token_info: Dict[str, Any] = {},
        **kwargs,
    ) -> bool:
        """Validate the token to see if it is a valid token from this generator.

        Args:
            token (str):
                The token to validate.

            token_info (dict, unused):
                A dictionary that contains information about the
                token.

            **kwargs (dict, unused):
                Additional keyword arguments.

        Returns:
            bool:
            Whether the token is a valid token from this generator.
        """
        return re.fullmatch(r'[0-9A-Fa-f]{40}', token) is not None



"""Field for atomically incrementing and decrementing counters in models."""

from django.db import models
from django.db.models import F
from django.db.models.expressions import Combinable as QueryExpressionType
from django.db.models.signals import post_init


class CounterField(models.IntegerField):
    """A field that provides atomic counter updating and smart initialization.

    The CounterField makes it easy to atomically update an integer,
    incrementing or decrementing it, without raise conditions or conflicts.
    It can update a single instance at a time, or a batch of objects at once.

    CounterField is useful for storing counts of objects, reducing the number
    of queries performed. This requires that the calling code properly
    increments or decrements at all the right times, of course.

    This takes an optional ``initializer`` parameter that, if provided, can
    be used to auto-populate the field the first time the model instance is
    loaded, perhaps based on querying a number of related objects. The value
    passed to ``initializer`` must be a function taking the model instance
    as a parameter, and must return an integer or ``None``. If it returns
    ``None``, the counter will not be updated or saved.

    The model instance will gain four new functions:

    :samp:`increment_{field_name}`
        Atomically increment the field by one.

    :samp:`decrement_{field_name}`
        Atomically decrement the field by one.

    :samp:`reload_{field_name}`
        Reload the value in this instance from the database.

    :samp:`reinit_{field_name}`
        Re-initializes the stored field using the initializer function.

    The field on the class (not the instance) provides two functions for
    batch-updating models:

    ``increment``
        Takes a queryset and increments this field for each object.

    ``decrement``
        Takes a queryset and decrements this field for each object.
    """

    @classmethod
    def increment_many(cls, model_instance, values, reload_object=True):
        """Increment several fields on a model instance at once.

        Args:
            model_instance (django.db.models.Model):
                The model instance containing the fields to increment.

            values (dict):
                A dictionary mapping field names to delta values to increment
                by.

            reload_object (bool, optional):
                Whether to reload the field values in ``model_instance`` to
                reflect the current values in the database after incrementing.
        """
        cls._update_values(model_instance, values, reload_object, multiplier=1)

    @classmethod
    def decrement_many(cls, model_instance, values, reload_object=True):
        """Decrement several fields on a model instance at once.

        Args:
            model_instance (django.db.models.Model):
                The model instance containing the fields to decrement.

            values (dict):
                A dictionary mapping field names to delta values to decrement
                by.

            reload_object (bool, optional):
                Whether to reload the field values in ``model_instance`` to
                reflect the current values in the database after decrementing.
        """
        cls._update_values(model_instance, values, reload_object,
                           multiplier=-1)

    @classmethod
    def _update_values(cls, model_instance, values, reload_object, multiplier):
        """Update several fields on a model instance at once.

        This does the hard work of updating fields for a model instance,
        updating existing values based on multiplying those provided with
        ``multiplier``.

        Args:
            model_instance (django.db.models.Model):
                The model instance containing the fields to update.

            values (dict):
                A dictionary mapping field names to delta values.

            reload_object (bool):
                Whether to reload the field values in ``model_instance`` to
                reflect the current values in the database after updating.

            multiplier (int):
                The value to multiply each of the provided values with.
        """
        cls._set_values(
            model_instance=model_instance,
            values={
                attname: F(attname) + value * multiplier
                for attname, value in values.items()
                if value != 0
            },
            reload_object=reload_object)

    @classmethod
    def _set_values(cls, model_instance, values, reload_object=True):
        """Set values for several fields on a model instance at once.

        Args:
            model_instance (django.db.models.Model):
                The model instance containing the fields to update.

            values (dict):
                A dictionary mapping field names to values. Each value can be
                anything allowed in a model update, including a
                :py:class:`query expression <django.db.models.F>`.

            reload_object (bool):
                Whether to reload the field values in ``model_instance`` to
                reflect the current values in the database after updating.
        """
        if not values:
            return

        model_cls = type(model_instance)
        model_cls.objects.filter(pk=model_instance.pk).update(**values)

        if reload_object:
            cls._reload_model_instance(model_instance, values.keys())

    @classmethod
    def _reload_model_instance(cls, model_instance, attnames):
        """Reload the value in this instance from the database.

        Args:
            model_instance (django.db.models.Model):
                The model instance containing fields to reload.

            attnames (list of str):
                The list of field attribute names to reload.
        """
        q = type(model_instance).objects.filter(pk=model_instance.pk)
        values = q.values(*attnames)[0]

        for attname, value in values.items():
            setattr(model_instance, attname, value)

    def __init__(self, verbose_name=None, name=None,
                 initializer=None, default=None, **kwargs):
        """Initialize the field.

        This can take a default value for the counter, or an initializer
        function that can compute a value (based on database queries or
        anything else needed).

        Args:
            verbose_name (unicode, optional):
                The verbose name to show users in forms. This defaults to a
                variation of the field name.

            name (str, optional):
                The name of the field. This defaults to the attribute name
                on the model.

            initializer (callable, optional):
                A function to call to compute an initial value for the field
                if ``None`` or when re-initializing.

            default (int, optional):
                An explicit default value for the field.

            **kwargs (dict):
                Additional keyword arguments for the field.
        """
        kwargs.update({
            'blank': True,
            'null': True,
        })

        super(CounterField, self).__init__(verbose_name, name, default=default,
                                           **kwargs)

        self._initializer = initializer
        self._locks = {}

    def increment(self, queryset, increment_by=1):
        """Increment this field on every object in the provided queryset.

        By default, this increments by 1, but a custom delta value can be
        provided.

        Args:
            queryset (django.db.models.query.QuerySet):
                The queryset for the update.

            increment_by (int, optional):
                The value to increment by. Defaults to 1.
        """
        queryset.update(**{self.attname: F(self.attname) + increment_by})

    def decrement(self, queryset, decrement_by=1):
        """Decrement this field on every object in the provided queryset.

        By default, this increments by 1, but a custom delta value can be
        provided.

        Args:
            queryset (django.db.models.query.QuerySet):
                The queryset for the update.

            decrement_by (int, optional):
                The value to decrement by. Defaults to 1.
        """
        queryset.update(**{self.attname: F(self.attname) - decrement_by})

    def contribute_to_class(self, cls, name):
        """Add methods to the model class.

        This introduces the methods for incrementing, decrementing,
        reloading, and re-initializing the fields.

        Args:
            cls (type):
                The model class.

            name (str):
                The name of the field on the model.
        """
        def _increment(model_instance, *args, **kwargs):
            self._increment(model_instance, *args, **kwargs)

        def _decrement(model_instance, *args, **kwargs):
            self._decrement(model_instance, *args, **kwargs)

        def _reload(model_instance):
            self._reload(model_instance)

        def _reinit(model_instance):
            self._reinit(model_instance)

        super(CounterField, self).contribute_to_class(cls, name)

        setattr(cls, 'increment_%s' % self.name, _increment)
        setattr(cls, 'decrement_%s' % self.name, _decrement)
        setattr(cls, 'reload_%s' % self.name, _reload)
        setattr(cls, 'reinit_%s' % self.name, _reinit)
        setattr(cls, self.attname, self)

        # We need to prevent CounterFields (or any subclasses) from being
        # saved by default. We'd ideally override save() or save_base(), but
        # can't safely do that here, and we can't modify update_fields due to
        # how it would impact the save process. So we need to go deeper,
        # overriding _do_update(), which actually receives the values and
        # field instances being saved. From there, we can filter out any
        # CounterFields by default.
        setattr(cls, '_do_update', type(self)._model_do_update)

        post_init.connect(self._post_init, sender=cls)

    def _increment(self, model_instance, reload_object=True, increment_by=1):
        """Increment this field by a value.

        By default, this increments by 1.

        Args:
            model_instance (django.db.models.Model):
                The model instance containing the field to increment.

            reload_object (bool, optional):
                Whether to reload the field value after incrementing.

            increment_by (int, optional):
                The value to increment by. Defaults to 1.
        """
        if increment_by != 0:
            cls = type(model_instance)
            self.increment(cls.objects.filter(pk=model_instance.pk),
                           increment_by)

            if reload_object:
                self._reload(model_instance)

    def _decrement(self, model_instance, reload_object=True, decrement_by=1):
        """Decrement this field by one.

        By default, this decrements by 1.

        Args:
            model_instance (django.db.models.Model):
                The model instance containing the field to decrement.

            reload_object (bool, optional):
                Whether to reload the field value after decrementing.

            increment_by (int, optional):
                The value to decrement by. Defaults to 1.
        """
        if decrement_by != 0:
            cls = type(model_instance)
            self.decrement(cls.objects.filter(pk=model_instance.pk),
                           decrement_by)

            if reload_object:
                self._reload(model_instance)

    def _reload(self, model_instance):
        """Reload the value in this instance from the database.

        Args:
            model_instance (django.db.models.Model):
                The model instance containing the field to reload.
        """
        self._reload_model_instance(model_instance, [self.attname])

    def _reinit(self, model_instance):
        """Re-initialize the value in the database from the initializer.

        Args:
            model_instance (django.db.models.Model):
                The model instance containing the field to re-initialize.
        """
        if not (model_instance.pk or self._initializer or
                callable(self._initializer)):
            # We don't want to end up defaulting this to 0 if creating a
            # new instance unless an initializer is provided. Instead,
            # we'll want to handle this the next time the object is
            # accessed.
            return

        value = 0

        if self._initializer:
            if isinstance(self._initializer, QueryExpressionType):
                value = self._initializer
            elif callable(self._initializer):
                model_instance_id = id(model_instance)
                self._locks[model_instance_id] = 1
                value = self._initializer(model_instance)
                del self._locks[model_instance_id]

        if value is not None:
            is_expr = isinstance(value, QueryExpressionType)

            if is_expr and not model_instance.pk:
                value = 0
                is_expr = False

            if is_expr:
                cls = model_instance.__class__
                cls.objects.filter(pk=model_instance.pk).update(**{
                    self.attname: value,
                })

                self._reload_model_instance(model_instance, [self.attname])
            else:
                setattr(model_instance, self.attname, value)

                if model_instance.pk:
                    model_instance.save(update_fields=[self.attname])

    def _post_init(self, instance, **kwargs):
        """Initialize the field when a model instance is created.

        This will begin the process of initializing the default value for
        the field. It checks to make sure there's not another initialization
        happening due to a second counter field on the same instance modifying
        the state.

        Args:
            instance (django.db.models.Model):
                The instance being initialized.

            **kwargs (dict):
                Extra keyword arguments for the signal handler.
        """
        # Prevent the possibility of recursive lookups where this
        # same CounterField on this same instance tries to initialize
        # more than once. In this case, this will have the updated
        # value shortly.
        if instance is not None:
            instance_id = id(instance)

            if instance_id not in self._locks:
                self._do_post_init(instance)

    def _do_post_init(self, instance):
        """Initialize the value for the field when a model instance is created.

        If a value doesn't already exist on the counter, and an initializer
        is provided, it will be called to set a new value for the counter.

        Args:
            instance (django.db.models.Model):
                The instance being initialized.
        """
        value = self.value_from_object(instance)

        if value is None:
            reinit = getattr(instance, 'reinit_%s' % self.name)
            reinit()

    @staticmethod
    def _model_do_update(model, base_qs, using, pk_val, values, update_fields,
                         forced_update):
        """Perform a SQL UPDATE for a table.

        We override Django's default implementation to first filter out some
        fields, allowing us to skip CounterFields (unless setting to None or
        explicitly requested in ``update_fields``).

        Args:
            model (django.db.models.Model):
                The model instance that this method was called on.

            base_qs (django.db.models.query.QuerySet):
                The queryset used for the update operation.

            using (unicode):
                The database connection being used.

            pk_val (int):
                The primary key value for the instance.

            values (list of tuple):
                The list of update values. Each item is a tuple in the form
                of ``(field_instance, None, value)``.

            update_fields (list of unicode):
                The specific fields the caller wanted to update. This may be
                ``None``.

            forced_update (bool):
                Whether the requester is specifically forcing a SQL UPDATE.

        Returns:
            bool:
            Whether there were any updates performed.
        """
        values = [
            value
            for value in values
            if (not isinstance(value[0], CounterField) or
                value[2] is None or
                (update_fields and value[0].attname in update_fields))
        ]

        # We explicitly need to use ``models.Model._do_update`` here instead of
        # calling ``super(type(model), model)._do_update``.
        #
        # If ``type(model)`` defines a counter field **and** inherits from an
        # abstract model that does the same, then
        # ``super(type(model), model)._do_update`` is **this method**. We will
        # infinitely recurse and blow the stack in this case. Hence we have to
        # skip right to ``models.Model``.
        return models.Model._do_update(
            model, base_qs, using, pk_val, values, update_fields,
            forced_update)



"""A flexible resource for the root of your API resource tree."""

import logging
from collections import defaultdict, namedtuple

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponseNotModified
from django.utils.translation import gettext_lazy as _
from django.urls import path

from djblets.urls.patterns import never_cache_patterns
from djblets.webapi.errors import DOES_NOT_EXIST
from djblets.webapi.resources.base import WebAPIResource
from djblets.webapi.responses import WebAPIResponseError


logger = logging.getLogger(__name__)


class RootResource(WebAPIResource):
    """The root of a resource tree.

    This is meant to be instantiated with a list of immediate child
    resources. The result of :py:meth:`get_url_patterns` should be included in
    a project's :file:`urls.py`.
    """

    name = 'root'
    singleton = True

    #: A resource entry returned from :py:meth:`RootResource.walk_resources`.
    #:
    #: Attributes:
    #:      name (str):
    #:          The name of current resource being explored.
    #:
    #       list_href (str):
    #:          The list_href associated with this resource.
    #:
    #       resource (djblets.webapi.resources.base.WebAPIResource):
    #:          The resource object being explored.
    #:
    #       is_list (bool):
    #:          Whether or not the resource is a list resource.
    #:
    #:      uri_template_name (str)
    #:          The name of the resource to use in the URI templates list.
    ResourceEntry = namedtuple('ResourceEntry',
                               ('name',
                                'list_href',
                                'resource',
                                'is_list',
                                'uri_template_name'))

    def __init__(self, child_resources=[], include_uri_templates=True):
        super(RootResource, self).__init__()
        self.list_child_resources = child_resources
        self._uri_templates = {}
        self._registered_uri_templates = {}
        self._include_uri_templates = include_uri_templates

    def get_etag(self, request, obj, *args, **kwargs):
        return self.encode_etag(request, repr(obj))

    def get(self, request, *args, **kwargs):
        """Retrieve the list of top-level resources and URL templates."""
        data = self.serialize_root(request, *args, **kwargs)
        etag = self.get_etag(request, data)

        if self.are_cache_headers_current(request, etag=etag):
            return HttpResponseNotModified()

        return 200, data, {
            'ETag': etag,
        }

    def serialize_root(self, request, *args, **kwargs):
        """Serialize the contents of the root resource.

        By default, this just provides links and URI templates. Subclasses
        can override this to provide additional data, or to otherwise
        change the structure of the root resource.
        """
        data = {
            'links': self.get_links(self.list_child_resources,
                                    request=request, *args, **kwargs),
        }

        if self._include_uri_templates:
            data['uri_templates'] = self.get_uri_templates(request, *args,
                                                           **kwargs)

        return data

    def get_uri_templates(self, request, *args, **kwargs):
        """Return all URI templates in the resource tree.

        REST APIs can be very chatty if a client wants to be well-behaved
        and crawl the resource tree asking for the links, instead of
        hard-coding the paths. The benefit is that they can keep from
        breaking when paths change. The downside is that it can take many
        HTTP requests to get the right resource.

        This list of all URI templates allows clients who know the resource
        name and the data they care about to simply plug them into the
        URI template instead of trying to crawl over the whole tree. This
        can make things far more efficient.

        Args:
            request (django.http.HttpRequest):
                The GET request for the Root resource.

            *args (tuple, unused):
                Additional unused arguments.

            **kwargs (dict, unused):
                Additional unused keyword arguments.

        Returns:
            dict:
                A mapping of resources to their URI templates.
        """
        if not self._uri_templates:
            self._uri_templates = {}

        base_href = request.build_absolute_uri()

        if base_href not in self._uri_templates:
            templates = defaultdict(list)
            unassigned_templates = self._registered_uri_templates.get(
                None, {})

            for uri_template_name, href in unassigned_templates.items():
                templates[uri_template_name].append(href)

            for entry in self.walk_resources(self, base_href):
                uri_template_name = entry.uri_template_name

                if uri_template_name is None:
                    continue

                templates[uri_template_name].append(entry.list_href)

                if entry.is_list:
                    list_templates = self._registered_uri_templates.get(
                        entry.resource, {})

                    for uri_template_name, href in list_templates.items():
                        templates[uri_template_name].append(
                            '%s%s' % (entry.list_href, href))

            final_templates = {}

            for uri_template_name, uri_template in templates.items():
                if len(uri_template) > 1:
                    if settings.DEBUG:
                        raise ImproperlyConfigured(
                            _('More than one URI template was mapped to '
                              'the "%(name)s" name: %(templates)s. Each URI '
                              'template must be mapped to a unique URI '
                              'template name in order to be included in the '
                              'URI templates list. This can be set through '
                              'the uri_template_name property.')
                            % {
                                'name': uri_template_name,
                                'templates': ', '.join(uri_template)
                            })
                    else:
                        logger.error(
                            'More than one URI template was mapped to '
                            'the "%s" name: %s. Only the first one will be '
                            'included in the URI templates list. To '
                            'include the other URI templates, they must '
                            'be mapped to a unique name by setting each '
                            'resource\'s uri_template_name property.'
                            % (uri_template_name, ', '.join(uri_template)))

                final_templates[uri_template_name] = uri_template[0]

            self._uri_templates[base_href] = final_templates

        return self._uri_templates[base_href]

    @classmethod
    def walk_resources(cls, resource, list_href):
        """Yield all URI endpoints associated with a specified resource.

        Args:
            resource djblets.webapi.resources.WebAPIResource:
                The starting point for searching the resource tree.

            list_href (unicode):
                The path to the list resource, relative to the WebAPIResource
                provided. Used as a component of the URL in the API.

        Yields:
            RootResource.ResourceEntry:
            Resource entries for all sub-resources.
        """
        yield cls.ResourceEntry(
            name=resource.name_plural,
            list_href=list_href,
            resource=resource,
            is_list=True,
            uri_template_name=resource.uri_template_name_plural)

        for child in resource.list_child_resources:
            child_href = '%s%s/' % (list_href, child.uri_name)

            for entry in cls.walk_resources(child, child_href):
                yield entry

        if resource.uri_object_key:
            object_href = '%s{%s}/' % (list_href, resource.uri_object_key)

            yield cls.ResourceEntry(
                name=resource.name,
                list_href=object_href,
                resource=resource,
                is_list=False,
                uri_template_name=resource.uri_template_name)

            for child in resource.item_child_resources:
                child_href = '%s%s/' % (object_href, child.uri_name)

                for entry in cls.walk_resources(child, child_href):
                    yield entry

    def api_404_handler(self, request, api_format=None, *args, **kwargs):
        """Default handler at the end of the URL patterns.

        This returns an API 404, instead of a normal django 404.
        """
        return WebAPIResponseError(request,
                                   err=DOES_NOT_EXIST,
                                   api_format=api_format)

    def get_url_patterns(self):
        """Return the Django URL patterns for this object and its children.

        This returns the same list as
        :py:meth:`WebAPIResource.get_url_patterns`, but also introduces a
        generic catch-all 404 handler which returns API errors instead of HTML.
        """
        urlpatterns = super(RootResource, self).get_url_patterns()
        urlpatterns += never_cache_patterns(
            path('<str>', self.api_404_handler))

        return urlpatterns

    def register_uri_template(self, name, relative_path,
                              relative_resource=None):
        """Register the specified resource for URI template serialization.

        This adds the specified name and relative resource to the Root
        Resource's URI templates.

        Args:
            name (unicode):
                The name of the associated resource being added to templates.

            relative_path (unicode):
                The path of the API resource relative to its parent resources.

            relative_resource (djblets.extensions.resources.ExtensionResource,
                               optional):
                The resource instance associated with this URI template.
        """
        # Clear the cache so that new lookups can detect newly added templates.
        self._uri_templates = {}

        templates = self._registered_uri_templates.setdefault(
            relative_resource, {})

        if name in templates:
            logger.debug(
                'The %s resource is already mapped to the following '
                'URI template: %s. This will be overwritten by the new URI '
                'template: %s.'
                % (name, templates[name], relative_path))

        # Allow the URI template to be overwritten if it already exists.
        templates[name] = relative_path

    def unregister_uri_template(self, name, relative_resource=None):
        """Unregister the specified resource for URI template serialization.

        This removes the specified name and relative resource to the Root
        Resource's URI templates.

        Args:
            name (unicode):
                The name of the resource being removed from templates.

            relative_resource (djblets.extensions.resources.ExtensionResource,
                               optional):
                The resource instance associated with this URI template.
        """
        # Clear the cache so that new lookups can detect newly added templates.
        self._uri_templates = {}

        try:
            del self._registered_uri_templates[relative_resource][name]
        except KeyError:
            pass



from collections import OrderedDict

import kgb
from django.contrib.auth.models import User
from django.db import models
from django.http import HttpResponseNotModified
from django.test.client import RequestFactory

from djblets.testing.testcases import TestCase, TestModelsLoaderMixin
from djblets.webapi.auth.backends import check_login
from djblets.webapi.fields import (ResourceFieldType,
                                   ResourceListFieldType,
                                   StringFieldType)
from djblets.webapi.resources.base import WebAPIResource
from djblets.webapi.resources.registry import (register_resource_for_model,
                                               unregister_resource_for_model,
                                               unregister_resource)
from djblets.webapi.responses import WebAPIResponse


class MyTestGroup(models.Model):
    name = models.CharField(max_length=10)
    users = models.ManyToManyField(User)


class BaseTestWebAPIResource(WebAPIResource):
    mimetype_vendor = 'djblets-test'


class MyTestUserResource(BaseTestWebAPIResource):
    name = 'user'
    model = User
    uri_object_key = 'pk'

    fields = {
        'username': {
            'type': StringFieldType,
        },
        'first_name': {
            'type': StringFieldType,
        },
        'last_name': {
            'type': StringFieldType,
        },
    }

    def get_href(self, obj, *args, **kwargs):
        return 'http://testserver/api/test/users/%s/' % obj.pk


class MyTestGroupResource(BaseTestWebAPIResource):
    name = 'group'
    model = MyTestGroup
    uri_object_key = 'pk'

    fields = {
        'name': {
            'type': StringFieldType,
        },
        'users': {
            'type': ResourceListFieldType,
        },
    }

    def get_href(self, obj, *args, **kwargs):
        return 'http://testserver/api/test/groups/%s/' % obj.pk

    def get_serializer_for_object(self, o):
        if isinstance(o, User):
            return MyTestUserResource()
        else:
            return self


class BaseTestRefUserResource(BaseTestWebAPIResource):
    name = 'test'
    uri_object_key = 'pk'

    def build_resource_url(self, name, request, **kwargs):
        return request.build_absolute_uri()

    def get_serializer_for_object(self, o):
        if isinstance(o, User):
            return MyTestUserResource()
        elif isinstance(o, MyTestGroup):
            return MyTestGroupResource()
        else:
            return self


class WebAPIResourceTests(kgb.SpyAgency, TestModelsLoaderMixin, TestCase):
    """Unit tests for djblets.webapi.resources.base."""

    tests_app = 'djblets.webapi.tests'

    def setUp(self):
        super(WebAPIResourceTests, self).setUp()

        self.maxDiff = 10000000

        self.factory = RequestFactory()
        self.test_resource = None

    def tearDown(self):
        super(WebAPIResourceTests, self).tearDown()

        if self.test_resource:
            unregister_resource(self.test_resource)

    def test_vendor_mimetypes(self):
        """Testing WebAPIResource with vendor-specific mimetypes"""
        class TestResource(WebAPIResource):
            mimetype_vendor = 'djblets'

        self.test_resource = TestResource()

        item_mimetypes = [
            mimetype['item']
            for mimetype in self.test_resource.allowed_mimetypes
            if 'item' in mimetype
        ]

        list_mimetypes = [
            mimetype['list']
            for mimetype in self.test_resource.allowed_mimetypes
            if 'list' in mimetype
        ]

        self.assertEqual(len(list_mimetypes), 4)
        self.assertEqual(len(item_mimetypes), 4)

        self.assertTrue('application/json' in
                        list_mimetypes)
        self.assertTrue('application/xml' in
                        list_mimetypes)
        self.assertTrue('application/vnd.djblets.testresources+json' in
                        list_mimetypes)
        self.assertTrue('application/vnd.djblets.testresources+xml' in
                        list_mimetypes)

        self.assertTrue('application/json' in
                        item_mimetypes)
        self.assertTrue('application/xml' in
                        item_mimetypes)
        self.assertTrue('application/vnd.djblets.testresource+json' in
                        item_mimetypes)
        self.assertTrue('application/vnd.djblets.testresource+xml' in
                        item_mimetypes)

    def test_vendor_mimetypes_with_custom(self):
        """Testing WebAPIResource with vendor-specific and custom mimetypes"""
        class TestResource(WebAPIResource):
            mimetype_vendor = 'djblets'
            allowed_mimetypes = WebAPIResource.allowed_mimetypes + [
                {'item': 'text/html'},
            ]

        self.test_resource = TestResource()

        item_mimetypes = [
            mimetype['item']
            for mimetype in self.test_resource.allowed_mimetypes
            if 'item' in mimetype
        ]

        list_mimetypes = [
            mimetype['list']
            for mimetype in self.test_resource.allowed_mimetypes
            if 'list' in mimetype
        ]

        self.assertEqual(len(list_mimetypes), 4)
        self.assertEqual(len(item_mimetypes), 5)

        self.assertTrue('application/json' in
                        list_mimetypes)
        self.assertTrue('application/xml' in
                        list_mimetypes)
        self.assertTrue('application/vnd.djblets.testresources+json' in
                        list_mimetypes)
        self.assertTrue('application/vnd.djblets.testresources+xml' in
                        list_mimetypes)

        self.assertTrue('application/json' in
                        item_mimetypes)
        self.assertTrue('application/xml' in
                        item_mimetypes)
        self.assertTrue('application/vnd.djblets.testresource+json' in
                        item_mimetypes)
        self.assertTrue('application/vnd.djblets.testresource+xml' in
                        item_mimetypes)
        self.assertTrue('application/vnd.djblets.testresource+xml' in
                        item_mimetypes)
        self.assertTrue('text/html' in
                        item_mimetypes)

    def test_get_with_vendor_mimetype(self):
        """Testing WebAPIResource with GET and vendor-specific mimetypes"""
        class TestResource(WebAPIResource):
            allowed_methods = ('GET', 'POST', 'PUT', 'DELETE')
            mimetype_vendor = 'djblets'
            uri_object_key = 'id'

            def get(self, *args, **kwargs):
                return 200, {}

            create = get
            update = get
            delete = get

        self.test_resource = TestResource()
        self._test_mimetype_responses(
            self.test_resource,
            '/api/tests/',
            'application/vnd.djblets.testresources+json',
            'application/vnd.djblets.testresources+xml')
        self._test_mimetype_responses(
            self.test_resource,
            '/api/tests/',
            'application/vnd.djblets.testresource+json',
            'application/vnd.djblets.testresource+xml',
            method='post')

        self._test_mimetype_responses(
            self.test_resource,
            '/api/tests/1/',
            'application/vnd.djblets.testresource+json',
            'application/vnd.djblets.testresource+xml',
            view_kwargs={'id': 1},
            method='put')
        self._test_mimetype_responses(
            self.test_resource,
            '/api/tests/1/',
            'application/vnd.djblets.testresource+json',
            'application/vnd.djblets.testresource+xml',
            view_kwargs={'id': 1},
            method='delete')

    def test_get_with_item_mimetype(self):
        """Testing WebAPIResource with GET and Item-Content-Type header"""
        class TestResource(WebAPIResource):
            allowed_methods = ('GET', 'POST', 'PUT', 'DELETE')
            mimetype_vendor = 'djblets'
            uri_object_key = 'id'

            def get(self, *args, **kwargs):
                return 200, {}

            create = get
            update = get
            delete = get

        self.test_resource = TestResource()
        self._test_item_mimetype_responses(
            self.test_resource,
            '/api/tests/',
            'application/vnd.djblets.testresources+json',
            'application/vnd.djblets.testresources+xml',
            'application/vnd.djblets.testresource+json',
            'application/vnd.djblets.testresource+xml')
        self._test_item_mimetype_responses(
            self.test_resource,
            '/api/tests/',
            'application/vnd.djblets.testresource+json',
            'application/vnd.djblets.testresource+xml',
            None,
            None,
            method='post')

        self._test_item_mimetype_responses(
            self.test_resource,
            '/api/tests/1/',
            'application/vnd.djblets.testresource+json',
            'application/vnd.djblets.testresource+xml',
            None,
            None,
            view_kwargs={'id': 1},
            method='put')
        self._test_item_mimetype_responses(
            self.test_resource,
            '/api/tests/',
            'application/vnd.djblets.testresources+json',
            'application/vnd.djblets.testresources+xml',
            'application/vnd.djblets.testresource+json',
            'application/vnd.djblets.testresource+xml')
        self._test_item_mimetype_responses(
            self.test_resource,
            '/api/tests/1/',
            'application/vnd.djblets.testresource+json',
            'application/vnd.djblets.testresource+xml',
            None,
            None,
            view_kwargs={'id': 1},
            method='delete')

    def test_get_with_autogenerate_etags_and_not_modified(self):
        """Testing WebAPIResource with GET and autogenerate_etags and requested
        ETag is a match
        """
        class TestResource(WebAPIResource):
            allowed_methods = ('GET', 'POST', 'PUT', 'DELETE')
            mimetype_vendor = 'djblets'
            uri_object_key = 'id'
            autogenerate_etags = True

            def get(self, *args, **kwargs):
                return 200, {
                    'abc': True,
                }

        resource = TestResource()
        etag = '790fc95c3f89afe28403c272553e790274ea1d3e'

        request = self.factory.get('/api/tests/',
                                   HTTP_ACCEPT='application/json',
                                   HTTP_IF_NONE_MATCH=etag)
        response = resource(request)

        self.assertIsInstance(response, HttpResponseNotModified)
        self.assertEqual(response.status_code, 304)
        self.assertEqual(response['ETag'], etag)

    def test_get_with_autogenerate_etags_and_modified(self):
        """Testing WebAPIResource with GET and autogenerate_etags and
        requested ETag is not a match
        """
        class TestResource(WebAPIResource):
            allowed_methods = ('GET', 'POST', 'PUT', 'DELETE')
            mimetype_vendor = 'djblets'
            uri_object_key = 'id'
            autogenerate_etags = True

            def get(self, *args, **kwargs):
                return 200, {
                    'abc': True,
                }

        resource = TestResource()
        request = self.factory.get('/api/tests/',
                                   HTTP_ACCEPT='application/json',
                                   HTTP_IF_NONE_MATCH='adfjhslfa')
        response = resource(request)

        self.assertIsInstance(response, WebAPIResponse)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response['ETag'],
                         '790fc95c3f89afe28403c272553e790274ea1d3e')

    def test_get_with_auth_response_headers(self):
        """Testing WebAPIResource with GET and ensuring that any headers
        passed from authentication backends appear in the response
        """
        class TestResource(WebAPIResource):
            allowed_methods = ('GET', 'POST', 'PUT', 'DELETE')
            mimetype_vendor = 'djblets'
            uri_object_key = 'id'
            autogenerate_etags = True

            def get(self, *args, **kwargs):
                return 200, {
                    'abc': True,
                }

        @self.spy_for(check_login)
        def _fake_check_login(request):
            return (True, None, {'X-test-header': 'Test'})

        resource = TestResource()
        request = self.factory.get('/api/tests/',
                                   HTTP_ACCEPT='application/json')
        response = resource(request)

        self.assertIsInstance(response, WebAPIResponse)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response['X-test-header'], 'Test')

    def test_are_cache_headers_current_with_old_last_modified(self):
        """Testing WebAPIResource.are_cache_headers_current with old last
        modified timestamp
        """
        request = RequestFactory().request()
        request.META['HTTP_IF_MODIFIED_SINCE'] = \
            'Wed, 14 Jan 2015 13:49:10 GMT'

        resource = WebAPIResource()
        self.assertFalse(resource.are_cache_headers_current(
            request, last_modified='Wed, 14 Jan 2015 12:10:13 GMT'))

    def test_are_cache_headers_current_with_current_last_modified(self):
        """Testing WebAPIResource.are_cache_headers_current with current last
        modified timestamp
        """
        timestamp = 'Wed, 14 Jan 2015 13:49:10 GMT'
        request = RequestFactory().request()
        request.META['HTTP_IF_MODIFIED_SINCE'] = timestamp

        resource = WebAPIResource()
        self.assertTrue(resource.are_cache_headers_current(
            request, last_modified=timestamp))

    def test_are_cache_headers_current_with_old_etag(self):
        """Testing WebAPIResource.are_cache_headers_current with old ETag"""
        request = RequestFactory().request()
        request.META['HTTP_IF_NONE_MATCH'] = 'abc123'

        resource = WebAPIResource()
        self.assertFalse(resource.are_cache_headers_current(request,
                                                            etag='def456'))

    def test_are_cache_headers_current_with_current_etag(self):
        """Testing WebAPIResource.are_cache_headers_current with current
        ETag
        """
        etag = 'abc123'
        request = RequestFactory().request()
        request.META['HTTP_IF_NONE_MATCH'] = etag

        resource = WebAPIResource()
        self.assertTrue(resource.are_cache_headers_current(request, etag=etag))

    def test_serialize_object_with_only_fields(self):
        """Testing WebAPIResource.serialize_object with
        ?only-fields=<fields>
        """
        class TestObject(object):
            field1 = 'abc'
            field2 = 'def'
            field3 = 'ghi'

        class TestResource(WebAPIResource):
            fields = {
                'field1': {
                    'type': StringFieldType,
                },
                'field2': {
                    'type': StringFieldType,
                },
                'field3': {
                    'type': StringFieldType,
                },
            }

        request = RequestFactory().get('/api/test/?only-fields=field1,field3')
        resource = TestResource()
        data = resource.serialize_object(TestObject(), request=request)

        self.assertEqual(data, {
            'field1': 'abc',
            'field3': 'ghi',
            'links': {
                'self': {
                    'href': 'http://testserver/api/test/'
                            '?only-fields=field1,field3',
                    'method': 'GET',
                },
            }
        })

    def test_serialize_object_with_only_fields_blank(self):
        """Testing WebAPIResource.serialize_object with ?only-fields="""
        class TestObject(object):
            field1 = 'abc'
            field2 = 'def'
            field3 = 'ghi'

        class TestResource(WebAPIResource):
            fields = {
                'field1': {
                    'type': StringFieldType,
                },
                'field2': {
                    'type': StringFieldType,
                },
                'field3': {
                    'type': StringFieldType,
                },
            }

        request = RequestFactory().get('/api/test/?only-fields=')
        resource = TestResource()
        data = resource.serialize_object(TestObject(), request=request)

        self.assertEqual(data, {
            'links': {
                'self': {
                    'href': 'http://testserver/api/test/?only-fields=',
                    'method': 'GET',
                },
            }
        })

    def test_serialize_object_with_expand_model(self):
        """Testing WebAPIResource.serialize_object with
        ?expand=<model_field>
        """
        class TestObject(object):
            pk = '1'
            user = User.objects.create(username='user1',
                                       first_name='Test1',
                                       last_name='User1')

        class TestResource(BaseTestRefUserResource):
            fields = {
                'user': {
                    'type': ResourceFieldType,
                    'resource': MyTestUserResource,
                },
            }

        request = RequestFactory().get('/api/test/?expand=user')
        resource = TestResource()
        data = resource.serialize_object(TestObject(), request=request)

        self.assertEqual(data, {
            '_expanded': {
                'user': {
                    'item_mimetype': 'application/vnd.djblets-test.user+json',
                },
            },
            'user': {
                'username': 'user1',
                'first_name': 'Test1',
                'last_name': 'User1',
                'links': {
                    'self': {
                        'href': 'http://testserver/api/test/users/1/',
                        'method': 'GET',
                    },
                },
            },
            'links': {
                'self': {
                    'href': 'http://testserver/api/test/?expand=user',
                    'method': 'GET',
                },
            }
        })

    def test_serialize_object_with_expand_queryset(self):
        """Testing WebAPIResource.serialize_object with
        ?expand=<queryset_field>
        """
        class TestObject(object):
            pk = '1'
            users = User.objects.all()

        class TestResource(BaseTestRefUserResource):
            fields = {
                'users': {
                    'type': ResourceListFieldType,
                    'resource': MyTestUserResource,
                },
            }

        User.objects.create(username='user1',
                            first_name='Test1',
                            last_name='User1')
        User.objects.create(username='user2',
                            first_name='Test2',
                            last_name='User2')

        request = RequestFactory().get('/api/test/?expand=users')
        resource = TestResource()
        data = resource.serialize_object(TestObject(), request=request)

        self.assertEqual(data, {
            '_expanded': {
                'users': {
                    'item_mimetype': 'application/vnd.djblets-test.user+json',
                },
            },
            'users': [
                {
                    'username': 'user1',
                    'first_name': 'Test1',
                    'last_name': 'User1',
                    'links': {
                        'self': {
                            'href': 'http://testserver/api/test/users/1/',
                            'method': 'GET',
                        },
                    },
                },
                {
                    'username': 'user2',
                    'first_name': 'Test2',
                    'last_name': 'User2',
                    'links': {
                        'self': {
                            'href': 'http://testserver/api/test/users/2/',
                            'method': 'GET',
                        },
                    },
                },
            ],
            'links': {
                'self': {
                    'href': 'http://testserver/api/test/?expand=users',
                    'method': 'GET',
                },
            }
        })

    def test_serialize_object_with_expand_child_resource(self):
        """Testing WebAPIResource.serialize_object with
        ?expand=<child_resource_field>
        """
        class TestObject(object):
            pk = '1'
            field = 'test'
            users = User.objects.all()

        class TestResource(BaseTestRefUserResource):
            item_child_resources = [MyTestUserResource()]

            fields = {
                'field': {
                    'type': StringFieldType,
                },
            }

        User.objects.create(username='user1',
                            first_name='Test1',
                            last_name='User1')
        User.objects.create(username='user2',
                            first_name='Test2',
                            last_name='User2')

        request = RequestFactory().get('/api/test/?expand=users')
        resource = TestResource()
        data = resource.serialize_object(TestObject(), request=request)

        self.assertEqual(data, {
            '_expanded': {
                'users': {
                    'item_mimetype': 'application/vnd.djblets-test.user+json',
                    'list_mimetype': 'application/vnd.djblets-test.users+json',
                    'list_url': 'http://testserver/api/test/users/',
                },
            },
            'field': 'test',
            'users': [
                {
                    'username': 'user1',
                    'first_name': 'Test1',
                    'last_name': 'User1',
                    'links': {
                        'self': {
                            'href': 'http://testserver/api/test/users/1/',
                            'method': 'GET',
                        },
                    },
                },
                {
                    'username': 'user2',
                    'first_name': 'Test2',
                    'last_name': 'User2',
                    'links': {
                        'self': {
                            'href': 'http://testserver/api/test/users/2/',
                            'method': 'GET',
                        },
                    },
                },
            ],
            'links': {
                'self': {
                    'href': 'http://testserver/api/test/?expand=users',
                    'method': 'GET',
                },
            }
        })

    def test_serialize_object_with_expand_in_parent_and_child_fields(self):
        """Testing WebAPIResource.serialize_object with
        ?expand= and field found in parent and expanded child resource
        """
        class TestObject(object):
            pk = '1'
            users = User.objects.all()
            groups = MyTestGroup.objects.all()

        class TestResource(BaseTestRefUserResource):
            item_child_resources = []

            fields = OrderedDict([
                ('groups', {
                    'type': ResourceListFieldType,
                }),
                ('users', {
                    'type': ResourceListFieldType,
                }),
            ])

        User.objects.create(username='user1',
                            first_name='Test1',
                            last_name='User1')
        user2 = User.objects.create(username='user2',
                                    first_name='Test2',
                                    last_name='User2')

        group = MyTestGroup.objects.create(name='group1')
        group.users.add(user2)

        request = RequestFactory().get('/api/test/?expand=users,groups')
        resource = TestResource()
        data = resource.serialize_object(TestObject(), request=request)

        self.assertEqual(data, {
            '_expanded': {
                'groups': {
                    'item_mimetype': 'application/vnd.djblets-test.group+json',
                },
                'users': {
                    'item_mimetype': 'application/vnd.djblets-test.user+json',
                },
            },
            'links': {
                'self': {
                    'href': 'http://testserver/api/test/?expand=users,groups',
                    'method': 'GET',
                },
            },
            'groups': [
                {
                    'name': 'group1',
                    'links': {
                        'self': {
                            'href': 'http://testserver/api/test/groups/1/',
                            'method': 'GET',
                        },
                    },
                    'users': [
                        {
                            'href': 'http://testserver/api/test/users/2/',
                            'method': 'GET',
                            'title': 'user2',
                        },
                    ],
                },
            ],
            'users': [
                {
                    'username': 'user1',
                    'first_name': 'Test1',
                    'last_name': 'User1',
                    'links': {
                        'self': {
                            'href': 'http://testserver/api/test/users/1/',
                            'method': 'GET',
                        },
                    },
                },
                {
                    'username': 'user2',
                    'first_name': 'Test2',
                    'last_name': 'User2',
                    'links': {
                        'self': {
                            'href': 'http://testserver/api/test/users/2/',
                            'method': 'GET',
                        },
                    },
                },
            ],
        })

    def test_serialize_object_with_only_fields_and_expand(self):
        """Testing WebAPIResource.serialize_object with
        ?only-fields=<field>&expand=<field>
        """
        class TestObject(object):
            pk = 1

            field1 = 'abc'
            field2 = User.objects.create(username='user1',
                                         first_name='Test1',
                                         last_name='Test1')
            field3 = User.objects.create(username='user2',
                                         first_name='Test2',
                                         last_name='Test2')

        class TestResource(BaseTestRefUserResource):
            fields = {
                'field1': {
                    'type': StringFieldType,
                },
                'field2': {
                    'type': ResourceFieldType,
                    'resource': MyTestUserResource,
                },
                'field3': {
                    'type': ResourceFieldType,
                    'resource': MyTestUserResource,
                },
            }

        request = RequestFactory().get(
            '/api/test/?only-fields=field2&expand=field2')
        resource = TestResource()
        obj = TestObject()
        data = resource.serialize_object(obj, request=request)

        self.assertEqual(data, {
            '_expanded': {
                'field2': {
                    'item_mimetype': 'application/vnd.djblets-test.user+json',
                },
            },
            'field2': {
                'links': {
                    'self': {
                        'href': 'http://testserver/api/test/users/1/',
                        'method': 'GET',
                    },
                },
            },
            'links': {
                'self': {
                    'href': 'http://testserver/api/test/?only-fields=field2'
                            '&expand=field2',
                    'method': 'GET',
                },
                'field3': {
                    'href': 'http://testserver/api/test/users/2/',
                    'method': 'GET',
                    'title': 'user2',
                },
            }
        })

    def test_serialize_object_with_only_links(self):
        """Testing WebAPIResource.serialize_object with ?only-links=<links>"""
        class TestObject(object):
            field1 = 'abc'
            field2 = 'def'
            field3 = 'ghi'

        class TestResource(WebAPIResource):
            allowed_methods = ('GET', 'POST', 'PUT', 'DELETE')
            fields = {
                'field1': {
                    'type': StringFieldType,
                },
                'field2': {
                    'type': StringFieldType,
                },
                'field3': {
                    'type': StringFieldType,
                },
            }

        request = RequestFactory().get('/api/test/?only-links=delete,update')
        resource = TestResource()
        data = resource.serialize_object(TestObject(), request=request)

        self.assertEqual(data, {
            'field1': 'abc',
            'field2': 'def',
            'field3': 'ghi',
            'links': {
                'delete': {
                    'href': 'http://testserver/api/test/',
                    'method': 'DELETE',
                },
                'update': {
                    'href': 'http://testserver/api/test/',
                    'method': 'PUT',
                },
            }
        })

    def test_serialize_object_with_only_links_blank(self):
        """Testing WebAPIResource.serialize_object with ?only-links="""
        class TestObject(object):
            field1 = 'abc'
            field2 = 'def'
            field3 = 'ghi'

        class TestResource(WebAPIResource):
            allowed_methods = ('GET', 'POST', 'PUT', 'DELETE')
            fields = {
                'field1': {
                    'type': StringFieldType,
                },
                'field2': {
                    'type': StringFieldType,
                },
                'field3': {
                    'type': StringFieldType,
                },
            }

        request = RequestFactory().get('/api/test/?only-links=')
        resource = TestResource()
        data = resource.serialize_object(TestObject(), request=request)

        self.assertEqual(data, {
            'field1': 'abc',
            'field2': 'def',
            'field3': 'ghi',
        })

    def test_serialize_object_with_cache_copy(self):
        """Testing WebAPIResource.serialize_object always returns a copy of
        the cached data
        """
        class TestObject(object):
            my_field = 'abc'

        request = RequestFactory().request()
        request.user = User()

        resource = WebAPIResource()
        resource.fields = {
            'my_field': {
                'type': StringFieldType,
            }
        }

        obj = TestObject()

        # We check this three times, since prior to Djblets 2.0.20, we would
        # first return a copy of the newly-generated data, then the cached
        # copy of the original data, and then the cached copy again (which
        # would no longer be untouched).
        data = resource.serialize_object(obj, request=request)
        self.assertIn('my_field', data)
        del data['my_field']

        data = resource.serialize_object(obj, request=request)
        self.assertIn('my_field', data)
        del data['my_field']

        data = resource.serialize_object(obj, request=request)
        self.assertIn('my_field', data)

    def test_uri_template_name_default(self):
        """Testing WebAPIResource.uri_template_name defaults to
        WebAPIResource.name
        """
        class TestResource(WebAPIResource):
            name = 'test'

        resource = TestResource()

        self.assertEqual(resource.uri_template_name, 'test')

    def test_uri_template_name_set(self):
        """Testing WebAPIResource.uri_template_name when set"""
        class TestResource(WebAPIResource):
            name = 'test'
            uri_template_name = 'template_name'

        resource = TestResource()

        self.assertEqual(resource.uri_template_name, 'template_name')

    def test_uri_template_name_plural_default(self):
        """Testing WebAPIResource.uri_template_name_plural defaults to
        WebAPIResource.name_plural
        """
        class TestResource(WebAPIResource):
            name = 'test'

        resource = TestResource()

        self.assertEqual(resource.uri_template_name_plural, 'tests')

    def test_uri_template_name_plural_set(self):
        """Testing WebAPIResource.uri_template_name_plural when set"""
        class TestResource(WebAPIResource):
            name = 'test'
            uri_template_name = 'name'
            uri_template_name_plural = 'plural_name'

        resource = TestResource()

        self.assertEqual(resource.uri_template_name_plural, 'plural_name')

    def test_uri_template_name_plural_none(self):
        """Testing WebAPIResource.uri_template_name_plural is None
        when WebAPIResource.uri_template_name is set to None
        """
        class TestResource(WebAPIResource):
            name = 'test'
            uri_template_name = None

        resource = TestResource()

        self.assertIsNone(resource.uri_template_name_plural)

    def test_uri_template_name_plural_singleton(self):
        """Testing WebAPIResource.uri_template_name_plural is
        WebAPIResource.uri_template_name for singletons
        """
        class TestResourceA(WebAPIResource):
            name = 'testA'
            uri_template_name = 'singleton'
            singleton = True

        class TestResourceB(WebAPIResource):
            name = 'testB'
            singleton = True

        resourceA = TestResourceA()
        resourceB = TestResourceB()

        self.assertEqual(resourceA.uri_template_name_plural, 'singleton')
        self.assertEqual(resourceB.uri_template_name_plural, 'testB')

    def test_uri_template_name_plural_with_uri_template_name_set(self):
        """Testing WebAPIResource.uri_template_name_plural is
        WebAPIResource.uri_template_name with an appended 's' for resources
        that have a uri_template_name set
        """
        class TestResource(WebAPIResource):
            name = 'test'
            uri_template_name = 'name'

        resource = TestResource()

        self.assertEqual(resource.uri_template_name_plural, 'names')

    def _test_mimetype_responses(self, resource, url, json_mimetype,
                                 xml_mimetype, **kwargs):
        self._test_mimetype_response(resource, url, '*/*', json_mimetype,
                                     **kwargs)
        self._test_mimetype_response(resource, url, 'application/json',
                                     json_mimetype, **kwargs)
        self._test_mimetype_response(resource, url, json_mimetype,
                                     json_mimetype, **kwargs)
        self._test_mimetype_response(resource, url, 'application/xml',
                                     xml_mimetype, **kwargs)
        self._test_mimetype_response(resource, url, xml_mimetype, xml_mimetype,
                                     **kwargs)

    def _test_mimetype_response(self, resource, url, accept_mimetype,
                                response_mimetype, method='get',
                                view_kwargs={}):
        func = getattr(self.factory, method)

        if accept_mimetype:
            request = func(url, HTTP_ACCEPT=accept_mimetype)
        else:
            request = func(url)

        response = resource(request, **view_kwargs)
        print(response)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response['Content-Type'], response_mimetype)

    def test_serialize_object_with_circular_references(self):
        """Testing WebAPIResource.serialize_object with circular references and
        ?expand=
        """
        class TestObject(models.Model):
            def __init__(self, name, pk):
                super(TestObject, self).__init__()

                self.name = name
                self.pk = pk  # Django 1.8+ requires a pk field

        class TestResource(BaseTestWebAPIResource):
            fields = {
                'dependency': {
                    'type': [TestObject],
                },
                'name': {
                    'type': StringFieldType,
                }
            }

        try:
            obj1 = TestObject('obj1', 1)
            obj2 = TestObject('obj2', 2)

            obj1.dependency = obj2
            obj2.dependency = obj1

            request = RequestFactory().get('/api/test/?expand=dependency')
            resource = TestResource()
            register_resource_for_model(TestObject, resource)
            data = resource.serialize_object(obj1, request=request)

            self.maxDiff = 100000
            self.assertEqual(data, {
                '_expanded': {
                    'dependency': {
                        'item_mimetype':
                            'application/vnd.djblets-test.testresource+json',
                    },
                },
                'dependency': {
                    'links': {
                        'dependency': {
                            'href': None,
                            'method': 'GET',
                            'title': 'TestObject object (1)',
                        },
                        'self': {
                            'href': 'http://testserver/api/test/?expand=dependency',
                            'method': 'GET',
                        },
                    },
                    'name': 'obj2',
                },
                'links': {
                    'self': {
                        'href': 'http://testserver/api/test/?expand=dependency',
                        'method': 'GET',
                    },
                },
                'name': 'obj1',
            })
        finally:
            unregister_resource_for_model(TestObject)

    def _test_item_mimetype_responses(self, resource, url, json_mimetype,
                                      xml_mimetype, json_item_mimetype,
                                      xml_item_mimetype, **kwargs):
        self._test_item_mimetype_response(resource, url, '*/*',
                                          json_item_mimetype, **kwargs)
        self._test_item_mimetype_response(resource, url, 'application/json',
                                          json_item_mimetype, **kwargs)
        self._test_item_mimetype_response(resource, url, json_mimetype,
                                          json_item_mimetype, **kwargs)
        self._test_item_mimetype_response(resource, url, 'application/xml',
                                          xml_item_mimetype, **kwargs)
        self._test_item_mimetype_response(resource, url, xml_mimetype,
                                          xml_item_mimetype, **kwargs)

    def _test_item_mimetype_response(self, resource, url, accept_mimetype,
                                     response_item_mimetype=None,
                                     method='get', view_kwargs={}):
        func = getattr(self.factory, method)

        if accept_mimetype:
            request = func(url, HTTP_ACCEPT=accept_mimetype)
        else:
            request = func(url)

        response = resource(request, **view_kwargs)
        print(response)
        self.assertEqual(response.status_code, 200)

        if response_item_mimetype:
            self.assertEqual(response['Item-Content-Type'],
                             response_item_mimetype)
        else:
            self.assertTrue('Item-Content-Type' not in response)
